/******************************************************************************
 *
 *  PalmOrb - Emulate a Matrix Orbital LCD Display with Keypad
 *  Copyright (C) 2004  Chris Lansley  <palmorb@lansley.me.uk>
 *  Copyright (C) 2001  David Douthitt 
 *  Copyright (C) 2000  J Robert Ray
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  $Source: /cvsroot/palmorb/palmorb/MatrixOrbital.cpp,v $
 *  $Revision: 1.11 $ $Date: 2004/11/02 21:03:46 $
 *
 ******************************************************************************/

        /************************************************************/
        /* This file contains mostly the Matrix Orbital emulation.  */
        /************************************************************/

#include <PalmOS.h>

#include "palmorb.h"
#include "screen.h"
#include "MatrixOrbital.h"
#include "serial.h"
#include "debug.h"
#include "segments.h"
#include "types.h"


#define CMD_CHAR 0xFE // 254 - command/escape character

#define MIN_CMD_CODE 35
#define MAX_CMD_CODE 147 
#define ARGCOUNT(x) (argCount[((x)-MIN_CMD_CODE)])
#define CMDFUNC(x) (cmdFunc[((x)-MIN_CMD_CODE)])
#define CMDNAME(x) (cmdName[((x)-MIN_CMD_CODE)])
#define MAX_CMD_LENGTH (80+2) // 254 64 & byte*80

#if DEBUG_ASSERTS
#define ASSERTXY(z) do{ ASSERT1(internalEmuState.xPos<internalEmuState.width,\
	   							z "ClipXY x>w: x=%u", internalEmuState.xPos);\
						ASSERT1(internalEmuState.yPos<internalEmuState.height,\
								z "ClipXY y>h: y=%u", internalEmuState.yPos);\
						} while(0)
#else // DEBUG_ASSERTS, !DEBUG_ASSERTS:
#define ASSERTXY(z) /*nothing*/
#endif // !DEBUG_ASSERTS

/* Exported Functions */
Err MO_Init(void) SEG2;
void MO_Fini(void) SEG2;
void MO_ProcessInput(UInt8 *chr, UInt32 size);
void MO_ClearOutputs(void);
void MO_ClearScreen(void);
void MO_SendKey (Char key);
void MO_SetState(UsedEmulationState *state);
void MO_GetState(UsedEmulationState *used,
	   		NotUsedEmulationState *notused,
			InternalEmulationState *internal);
void MO_enterParsedMode(void);
void MO_enterHexMode(void);
void MO_returnToNormalMode(void);
void MO_PrefsChanged(void);


/* Exported variables */
UInt32 MO_ProtocolErrors=0;


/* File Global variables */
typedef Int32 (*cmd_func)(UInt8 *);
static cmd_func cmdFunc[MAX_CMD_CODE-MIN_CMD_CODE+1];
static UInt8 argCount[MAX_CMD_CODE-MIN_CMD_CODE+1];
static Char *cmdName[MAX_CMD_CODE-MIN_CMD_CODE+1];

static Char keyBuf[10]; 
static UInt8 keyBufLen;
static UInt8 mustAutoscroll;
static Boolean backlightTrack;

static InternalEmulationState internalEmuState;
static NotUsedEmulationState notUsedEmuState;
static UsedEmulationState    usedEmuState;
static Boolean showHex = false;
static Boolean showCmds = false;

/* Local Functions */
static void updateBacklight(void);
static void init_argCounts(void) SEG2;
static void init_cmdFuncs(void) SEG2;
static void init_cmdNames(void) SEG2;
static void retreatPos (void);
static void advancePos (void);
static Boolean clipXY (void);
static void putOutChar(Char c);
static void goToXY(UInt8 x, UInt8 y);
static void sendSerialNum(void);
static void putLargeNum (Char *str,  UInt16 x);
static void displayBigNum(UInt16 num, UInt16 x);
static void drawVBar(UInt16 col, UInt16 barlen);
static void drawHorizontalBar(UInt16 dir, UInt16 barlen);
static void doLineFeed (void);
static void doCarriageReturn (void);
static void debug_startOnNewline (void);
static void debug_outputString (Char *str);
static void debug_outputHex(UInt8 byte);
static void debug_outputChar (UInt8 chr);
static void debug_hexOutput (UInt8 *buf, UInt32 size);
static void debug_displayFunc(UInt8 *buf);
static void debug_displayBadFunc(UInt8 cmd);

     /*Escape functions*/
static Int32 escape_displayBigNum(UInt8 *);
static Int32 escape_GetFWVersion(UInt8 *);
static Int32 escape_getKeyBuf(UInt8 *);
static Int32 escape_ClearKeyBuf(UInt8 *);
static Int32 escape_retreatPos(UInt8 *);
static Int32 escape_advancePos(UInt8 *);
static Int32 escape_ClearScreen(UInt8 *);
static Int32 escape_ignore(UInt8 *);
static Int32 escape_BinaryCharacter(UInt8 *);
static Int32 escape_GetModuleType(UInt8 *);
static Int32 escape_SetDebounceTime(UInt8 *);
static Int32 escape_SetBaud(UInt8 *);
static Int32 escape_OutputSerialNo(UInt8 *);
static Int32 escape_SetSerialNo(UInt8 *);
static Int32 escape_SetLinewrapOn(UInt8 *);
static Int32 escape_SetLinewrapOff(UInt8 *);
static Int32 escape_SetAutoscrollOn(UInt8 *);
static Int32 escape_SetAutoscrollOff(UInt8 *);
static Int32 escape_SetAutotransmitOn(UInt8 *);
static Int32 escape_SetAutotransmitOff(UInt8 *);
static Int32 escape_SetBlockOn(UInt8 *);
static Int32 escape_SetBlockOff(UInt8 *);
static Int32 escape_SetUnderlineOn(UInt8 *);
static Int32 escape_SetUnderlineOff(UInt8 *);
static Int32 escape_SetAutorepeatOn(UInt8 *);
static Int32 escape_SetAutorepeatOff(UInt8 *);
static Int32 escape_Home(UInt8 *);
static Int32 escape_SetBacklightOn(UInt8 *);
static Int32 escape_SetBacklightOff(UInt8 *);
static Int32 escape_SetFlowControlOn(UInt8 *);
static Int32 escape_SetFlowControlOff(UInt8 *);
static Int32 escape_MoveToXY(UInt8 *);
static Int32 escape_SetI2CAddress(UInt8 *);
static Int32 escape_SetContrast(UInt8 *);
static Int32 escape_SetOutputOn(UInt8 *);
static Int32 escape_SetOutputOff(UInt8 *);
static Int32 escape_SetStartupScreen(UInt8 *);
static Int32 escape_InitVBarNarrow(UInt8 *);
static Int32 escape_InitVBarWide(UInt8 *);
static Int32 escape_InitHBar(UInt8 *);
static Int32 escape_InitBigNum(UInt8 *);
static Int32 escape_drawVBar(UInt8 *);
static Int32 escape_DrawHBar(UInt8 *);


Err 
MO_Init(UsedEmulationState *defaults) //SEG2
{
	init_argCounts();
	init_cmdFuncs();
	init_cmdNames();

	internalEmuState.numOutputs = 6;
	internalEmuState.numKeys = 25;
	internalEmuState.width = 20;
	internalEmuState.height = 4; //MAX15, 4 == normal
	internalEmuState.xPos = 0;
	internalEmuState.yPos = 0;
	keyBufLen=0;
	mustAutoscroll=0;

	notUsedEmuState.underline=0;
	notUsedEmuState.block=0;
	notUsedEmuState.autorepeat=0;
	notUsedEmuState.autorepeatMode=0;
	notUsedEmuState.contrast=128;
	notUsedEmuState.debounceTime=52;
	notUsedEmuState.addressI2C=0x50;

	MO_SetState(defaults);
	return errNone;
}

void
MO_enterHexMode(void)
{
	MO_ClearScreen();
	showHex = true;
	showCmds= false;
	usedEmuState.autoscroll=true;
	usedEmuState.linewrap=true;
}

void
MO_enterParsedMode(void)
{
	MO_ClearScreen();
	showHex = false;
	showCmds= true;
	usedEmuState.autoscroll=true;
	usedEmuState.linewrap=true;
}

void
MO_returnToNormalMode(void)
{
	MO_ClearScreen();
	showHex = false;
	showCmds= false;
}


void
MO_Fini(void) //SEG2
{
	// nothing to do here - yet.
}

void
MO_PrefsChanged(void)
{
	updateBacklight();
}

static void 
updateBacklight(void)
{
	if (!AppPrefs.backlight && backlightTrack)
	{ // If we aren't allow to control the light and we turned it on
	  // turn it back off.
		backlightTrack=0;
		EvtEnqueueKey(vchrBacklight, 0, commandKeyMask); 
	}

	if (AppPrefs.backlight)
	{ // If we are allowed to control the light...
		if (usedEmuState.backlight != backlightTrack)
		{
			EvtEnqueueKey(vchrBacklight, 0, commandKeyMask); 
			backlightTrack = usedEmuState.backlight;
		}
	}
}


/* Sets the device to the given defauts.
 * Should be called after MO_Init, and may be
 * called again at any time (but before MO_Fini).
 */
void 
MO_SetState(UsedEmulationState *state)
{
	usedEmuState = *state; //structure copy.

	// prepare for when screen size is in usedEmuState
	Screen_SetConfig(internalEmuState.width, internalEmuState.height,
		   	internalEmuState.numOutputs, internalEmuState.numKeys);

	updateBacklight();
}

void 
MO_GetState(UsedEmulationState *used,
	   		NotUsedEmulationState *notused,
			InternalEmulationState *internal)
{
	if (used != NULL)
		*used = usedEmuState;	
	if (notused != NULL)
		*notused = notUsedEmuState;	
	if (internal != NULL)
		*internal = internalEmuState;	
}

static void 
init_argCounts(void) // SEG2
{
	ARGCOUNT(35) = 2;
	ARGCOUNT(51) = 1;
	ARGCOUNT(52) = 2; //CJL: was using 4!
	ARGCOUNT(57) = 1;
	ARGCOUNT(58) = 2;
	ARGCOUNT(61) = 2;
	ARGCOUNT(64) = 80; //CJL: depends on screen size...
	ARGCOUNT(66) = 1;
	ARGCOUNT(71) = 2;
	ARGCOUNT(78) = 9;
	ARGCOUNT(80) = 1;
	ARGCOUNT(85) = 1;
	ARGCOUNT(86) = 1;
	ARGCOUNT(87) = 1;
	ARGCOUNT(124) = 4;
	ARGCOUNT(126) = 1;
	ARGCOUNT(145) = 1;
	ARGCOUNT(147) = 1;
}

static void 
init_cmdFuncs(void) //SEG2
{
	CMDFUNC(35) = escape_displayBigNum;
	CMDFUNC(38) = escape_getKeyBuf;
	CMDFUNC(51) = escape_SetI2CAddress;
	CMDFUNC(52) = escape_SetSerialNo;
	CMDFUNC(53) = escape_OutputSerialNo;
	CMDFUNC(54) = escape_GetFWVersion;
	CMDFUNC(55) = escape_GetModuleType; 
	CMDFUNC(57) = escape_SetBaud;
	CMDFUNC(58) = escape_SetFlowControlOn;
	CMDFUNC(59) = escape_SetFlowControlOff;
	CMDFUNC(61) = escape_drawVBar;
	CMDFUNC(64) = escape_SetStartupScreen;
	CMDFUNC(65) = escape_SetAutotransmitOn;
	CMDFUNC(66) = escape_SetBacklightOn;
	CMDFUNC(67) = escape_SetLinewrapOn;
	CMDFUNC(68) = escape_SetLinewrapOff;
	CMDFUNC(69) = escape_ClearKeyBuf;
	CMDFUNC(70) = escape_SetBacklightOff;
	CMDFUNC(71) = escape_MoveToXY;
	CMDFUNC(72) = escape_Home;
	CMDFUNC(74) = escape_SetUnderlineOn;
	CMDFUNC(75) = escape_SetUnderlineOff;
	CMDFUNC(76) = escape_retreatPos;
	CMDFUNC(77) = escape_advancePos;
	CMDFUNC(78) = escape_BinaryCharacter;
	CMDFUNC(79) = escape_SetAutotransmitOff;
	CMDFUNC(80) = escape_SetContrast;
	CMDFUNC(81) = escape_SetAutoscrollOn;
	CMDFUNC(82) = escape_SetAutoscrollOff;
	CMDFUNC(83) = escape_SetBlockOn;
	CMDFUNC(84) = escape_SetBlockOff;
	CMDFUNC(85) = escape_SetDebounceTime;
	CMDFUNC(86) = escape_SetOutputOff;
	CMDFUNC(87) = escape_SetOutputOn;
	CMDFUNC(88) = escape_ClearScreen;
	CMDFUNC(96) = escape_SetAutorepeatOff;
	CMDFUNC(104)= escape_InitHBar;
	CMDFUNC(110)= escape_InitBigNum;
	CMDFUNC(115)= escape_InitVBarNarrow;
	CMDFUNC(118)= escape_InitVBarWide;
	CMDFUNC(124)= escape_DrawHBar;
	CMDFUNC(126)= escape_SetAutorepeatOn;
	CMDFUNC(145)= escape_SetContrast;
	/*---------------------------------------------*
	 * COMMANDS TO IGNORE....
	 *---------------------------------------------*/
	CMDFUNC(147)= escape_ignore;//254 147 <0|1>: remember settings
}

static void 
init_cmdNames(void) // SEG2
{
	CMDNAME(35) = "displayBigNum";
	CMDNAME(38) = "GetKeyBuf";
	CMDNAME(51) = "SetI2CAddress";
	CMDNAME(52) = "SetSerialNo";
	CMDNAME(53) = "OutputSerialNo";
	CMDNAME(54) = "GetFirmwareVer";
	CMDNAME(55) = "GetModuleType"; 
	CMDNAME(57) = "SetBaud";
	CMDNAME(58) = "SetFlowControlOn";
	CMDNAME(59) = "SetFlowControlOff";
	CMDNAME(61) = "drawVBar";
	CMDNAME(64) = "SetStartupScreen";
	CMDNAME(65) = "SetAutotransmitOn";
	CMDNAME(66) = "SetBacklightOn";
	CMDNAME(67) = "SetLinewrapOn";
	CMDNAME(68) = "SetLinewrapOff";
	CMDNAME(69) = "ClearKeyBuf";
	CMDNAME(70) = "SetBacklightOff";
	CMDNAME(71) = "MoveToXY";
	CMDNAME(72) = "Home";
	CMDNAME(74) = "SetUnderlineOn";
	CMDNAME(75) = "SetUnderlineOff";
	CMDNAME(76) = "Pos--";
	CMDNAME(77) = "Pos++";
	CMDNAME(78) = "DefineChar";
	CMDNAME(79) = "SetAutotransmitOff";
	CMDNAME(80) = "SetContrast";
	CMDNAME(81) = "SetAutoscrollOn";
	CMDNAME(82) = "SetAutoscrollOff";
	CMDNAME(83) = "SetBlockOn";
	CMDNAME(84) = "SetBlockOff";
	CMDNAME(85) = "SetDebounceTime";
	CMDNAME(86) = "SetOutputOff";
	CMDNAME(87) = "SetOutputOn";
	CMDNAME(88) = "Clear";
	CMDNAME(96) = "SetAutorepeatOff";
	CMDNAME(104)= "InitHBar";
	CMDNAME(110)= "InitBigNum";
	CMDNAME(115)= "InitVBarNarrow";
	CMDNAME(118)= "InitVBarWide";
	CMDNAME(124)= "DrawHBar";
	CMDNAME(126)= "SetAutorepeatOn";
	CMDNAME(145)= "SetContrast";
	/* these are completely ignored*/
	CMDNAME(147)= "Remember";
}



static void
doLineFeed (void) 
{
	internalEmuState.yPos++;
	if (internalEmuState.yPos > internalEmuState.height - 1) 
	{
		if (usedEmuState.autoscroll == 1) 
		{
			Screen_DoAutoscroll(); // do it right now... for LF..
			internalEmuState.yPos = internalEmuState.height - 1;
		}
		else
			internalEmuState.yPos = 0;
	}
}

static void
doCarriageReturn (void) 
{
	internalEmuState.xPos = 0;
}

void 
MO_SendKey (Char key)
{
	if (usedEmuState.autotransmit)
	{
		Serial_SendChar(key);
		// Send a second one to act as key up.
		if (AppPrefs.sendKeyUpDown)
			Serial_SendChar(key);
	}
	else
	{
		if (keyBufLen<sizeof(keyBuf))
		{
			keyBuf[keyBufLen]=key;
			keyBufLen++;
		}
	}
}

void
MO_ClearScreen(void)
{
	Screen_Clear();
	mustAutoscroll = 0;
	internalEmuState.xPos = 0;
	internalEmuState.yPos = 0;
}

static Boolean
clipXY (void) 
{
	// don't check xPos and yPos going below 0 - they're unsigned.
	Boolean clipped = false;

	if (internalEmuState.xPos >= internalEmuState.width) 
	{
		PROTOASSERT(internalEmuState.xPos<internalEmuState.width, "clipXY x>w");
		internalEmuState.xPos = internalEmuState.width - 1;
		clipped=true;
	}

	if (internalEmuState.yPos >= internalEmuState.height) 
	{
		PROTOASSERT(internalEmuState.yPos<internalEmuState.height, "clipXY y>h");
		internalEmuState.yPos = internalEmuState.height - 1;
		clipped=true;
	} 

	return(clipped);
}

static void
advancePos (void) 
{
	internalEmuState.xPos++;

	// check if x goes past maximum value...
	if (internalEmuState.xPos >= internalEmuState.width) 
	{
		internalEmuState.xPos = 0;

		// check for linewrap...
		if (usedEmuState.linewrap == 1) 
		{
			/*
			 * This is like the linefeed code - except it doesn't do the
			 * LineFeed until more data arrives.
			 */
			internalEmuState.yPos++; // advance to next line down...

			// check if y goes beyond maximum...
			if (internalEmuState.yPos >= internalEmuState.height) 
			{
				// check for autoscroll..
				if (usedEmuState.autoscroll == 1) 
				{
					mustAutoscroll = 1;  // move screen up one line...
					// set y to maximum value...
					internalEmuState.yPos = internalEmuState.height - 1; 
				}
				else
					internalEmuState.yPos = 0;
			}
		}
	}
}

static void
retreatPos (void) 
{
	// Got to be careful here as xPos and yPos are unsigned!

	if (internalEmuState.xPos == 0) 
	{
		// set x to maximum
		internalEmuState.xPos = internalEmuState.width - 1; 

		// check for linewrap
		if (usedEmuState.linewrap == 1) 
		{

			// check if y goes beyond minimum value...
			if (internalEmuState.yPos == 0)
			{
				// set y to maximum
				internalEmuState.yPos = internalEmuState.height - 1; 
			}
			else
			{
				internalEmuState.yPos--; // go back up one line...
			}
		}
	}
	else
	{
		internalEmuState.xPos--;
	}
}

void
MO_ClearOutputs(void)
{
    for (UInt32 i = 1; i <= internalEmuState.numOutputs; i++)
        Screen_setOutput(i, 0);
}



static void 
putOutChar(Char c)
{
	if (mustAutoscroll)
	{
		Screen_DoAutoscroll();
		mustAutoscroll=0;
	}

	ASSERTXY("putOutChar: ");

	Screen_PutChar(c, internalEmuState.xPos, internalEmuState.yPos);

	advancePos(); // handles linewrap and autoscroll too
}




// Expects 0 based coords
static void
goToXY(UInt8 x, UInt8 y) 
{
	Boolean clipped;

	internalEmuState.xPos = x;
	internalEmuState.yPos = y;
	clipped=clipXY();

	if (clipped)
		MO_ProtocolErrors++;
}


static void
putLargeNum (Char *str,  UInt16 x) 
{
	for (UInt32 i=internalEmuState.height; i > 0; i--) 
	{
		goToXY(x, internalEmuState.height-i);
		Screen_PutNCharsAt(str, 3, x, internalEmuState.height-i);
		str += 3;
	}
}

#define FULL_BLOCK	'\xFF'
#define BLANK		' '
#define BOTTOM_HALF	'\x1'
#define TOP_HALF	'\x0'
#define LEFT_HALF	'\x2'
#define RIGHT_HALF	'\x3'
#define TOP_RIGHT	'\x6'
#define TOP_LEFT	'\x4'
#define BOTTOM_RIGHT	'\x5'
#define BOTTOM_LEFT	'\x7'

static void
displayBigNum(UInt16 num, UInt16 x) 
{
	Char Digit[10][12] = {
				// Number 0

				{
				BOTTOM_RIGHT,	FULL_BLOCK,	BOTTOM_LEFT,
				FULL_BLOCK,	BLANK,		FULL_BLOCK,
				FULL_BLOCK,	BLANK,		FULL_BLOCK,
				TOP_RIGHT,	FULL_BLOCK,	TOP_LEFT
				},

				// Number 1

				{
				BLANK,		BOTTOM_RIGHT,	FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK
				},

				// Number 2

				{
				BOTTOM_RIGHT,	FULL_BLOCK,	BOTTOM_LEFT,
				BLANK,		BOTTOM_RIGHT,	TOP_LEFT,
				BOTTOM_RIGHT,	TOP_LEFT,	BLANK,
				FULL_BLOCK,	FULL_BLOCK,	FULL_BLOCK
				},

				// Number 3

				{
				BOTTOM_RIGHT,	FULL_BLOCK,	BOTTOM_LEFT,
				BLANK,		BOTTOM_RIGHT,	TOP_LEFT,
				BLANK,		TOP_RIGHT,	BOTTOM_LEFT,
				TOP_RIGHT,	FULL_BLOCK,	TOP_LEFT
				},

				// Number 4

				{
				FULL_BLOCK,	BLANK,		FULL_BLOCK,
				FULL_BLOCK,	BOTTOM_HALF,	FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK
				},

				// Number 5

				{
				FULL_BLOCK,	FULL_BLOCK,	FULL_BLOCK,
				FULL_BLOCK,	BLANK,		BLANK,
				TOP_HALF,	TOP_HALF,	BOTTOM_LEFT,
				BOTTOM_HALF,	BOTTOM_HALF,	TOP_LEFT
				},

				// Number 6

				{
				BOTTOM_RIGHT,	FULL_BLOCK,		BOTTOM_LEFT,
				FULL_BLOCK,		BLANK,			BLANK,
				FULL_BLOCK,		TOP_HALF,		BOTTOM_LEFT,
				TOP_RIGHT,		FULL_BLOCK,		TOP_LEFT
				},

				// Number 7

				{
				FULL_BLOCK,	FULL_BLOCK,	FULL_BLOCK,
				BLANK,		BOTTOM_RIGHT,	TOP_LEFT,
				BOTTOM_RIGHT,	TOP_LEFT,	BLANK,
				FULL_BLOCK,	BLANK,		BLANK
				},

				// Number 8

				{
				BOTTOM_RIGHT,	FULL_BLOCK,	BOTTOM_LEFT,
				FULL_BLOCK,	BOTTOM_HALF,	FULL_BLOCK,
				FULL_BLOCK,	BLANK,		FULL_BLOCK,
				TOP_RIGHT,	FULL_BLOCK,	TOP_LEFT
				},

				// Number 9

				{
				BOTTOM_RIGHT,	FULL_BLOCK,	BOTTOM_LEFT,
				TOP_RIGHT,	BOTTOM_HALF,	FULL_BLOCK,
				BLANK,		BLANK,		FULL_BLOCK,
				TOP_RIGHT,	FULL_BLOCK,	TOP_LEFT
				}

			};

	ASSERT((num<=9) && (num>=0), "bignum out of range");
	if ((num > 9) || (num < 0))
		return;

	putLargeNum(Digit[num], x);
}

/* Escape Function
 * 		254 54
 * 		(Esc 6)
 * 	Read Firmware version number
 */
static Int32
escape_GetFWVersion(UInt8 *ignored)
{
	ASSERT(ignored[0]==54, "wrong func");
	// Not sure what to output - assuming this is used
	// to avoid bugs in earlier firmwares then we might
	// as well as high as we can. But lets avoid sign
	// problems in the client.
	Serial_SendChar(0x7F); //Assuming it only wants one byte
	return 0;
}

/* Escape Function
 *		254 76  
 *		(Esc L)
 * move cursor one pos to left
 */
static Int32 
escape_retreatPos(UInt8 *ignored)
{
	ASSERT(ignored[0]==76, "wrong func");

	retreatPos();
	return(0);
}

/* Escape Function
 * 		254 69
 * 		(Esc E)
 * 	Clear key buffer
 */
static Int32
escape_ClearKeyBuf(UInt8 *ignored)
{
	ASSERT(ignored[0]==69, "wrong func");
	PROTOASSERT(!usedEmuState.autotransmit, "autotransmit but ClearKeybuf?");

	keyBufLen=0;

	return(0);
}

/* Escape Function
 *		254 38
 *		(Esc &)
 * get buffered key presses - or 0x00 if empty
 * High order bit set unless this is the last/only key press.
 * 
 * Just return 0x00.
 */
static Int32
escape_getKeyBuf(UInt8 *ignored)
{
	ASSERT(ignored[0]==38, "wrong func");

	PROTOASSERT(!usedEmuState.autotransmit, "autotransmit but getKeybuf?");

	if (usedEmuState.autotransmit)
		return(0);

	if (keyBufLen==0)
	{
		Serial_SendChar(0x00);
	}
	else
	{
		// send all but the last key
		for (UInt32 i=0; i<keyBufLen-1U; i++)
		{
			Serial_SendChar(keyBuf[i]|0x80); // Set top bit - "more data coming"
		}
		// now send the last key (without top bit set)
		Serial_SendChar(keyBuf[keyBufLen-1]);
		keyBufLen=0;
	}
	return(0);
}

/* Escape Function
 * 		254 77  
 * 		(Esc M)
 * move cursor one pos to right
 */
static Int32 
escape_advancePos(UInt8 *ignored)
{
	ASSERT(ignored[0]==77, "wrong func");

	advancePos();
	return(0);
}

/* Escape Function
 * 		254 88
 * 		(Esc X)
 * Clear screen...
 */
static Int32 
escape_ClearScreen(UInt8 *ignored)
{
	ASSERT(ignored[0]==88, "wrong func");

	MO_ClearScreen();
	return(0);
}

/* Escape Function
 * 		
 * Ignore the command
 */
static Int32 
escape_ignore(UInt8 *ignored)
{
	return(0);
}

/* Escape Function
 * 		254 78 <c> <byte>*8
 * 		(Esc N)
 * define (5x8) char <c> using binary data <b>
 * <c> is 0..7
 */
static Int32 
escape_BinaryCharacter (UInt8 *args) 
{
	ASSERT(args[0]==78, "wrong func");
	if (args[1]>=NUM_SPECIAL_CHARS)
	{
		PROTOASSERT(args[1]<NUM_SPECIAL_CHARS, "special char out of range");
		MO_ProtocolErrors++;
		return(0);
	}

	Screen_DefineChar(args[1], args+2);
	return(0);
}


/* Escape Function
 * 		254 55
 * 		(Esc 7)
 * Read Module type
 */
static Int32 
escape_GetModuleType(UInt8 *ignored)
{
	ASSERT(ignored[0]==55, "wrong func");

	Serial_SendChar(9); // 9 == LK204-25
	return(0);
}


/* Escape Function
 * 		254 85 <time>
 * 		(Esc U)
 * Set Debounce Time
 */
static Int32 
escape_SetDebounceTime(UInt8 *args)
{
	ASSERT(args[0]==85, "wrong func");

	notUsedEmuState.debounceTime = args[1];
	return (0);
}


/* Escape Function
 * 		254 57 <speed>
 * 		(Esc 9)
 * Set baud rate...
 */
static Int32 
escape_SetBaud(UInt8 *args) 
{
	UInt32 baudrate=0;

	ASSERT(args[0]==57, "wrong func");

	switch (args[1]) 
	{
		case 0xFF: baudrate = 1200; break;
		case 0x81: baudrate = 2400; break;
		case 0x20: baudrate = 9600; break;
		case 0x0F: baudrate = 19200; break;
		case 0x95: baudrate = 57600; break;
		case 0x03: baudrate = 76800; break;
		case 0x8A: baudrate = 115000; break;
		default:
			   PROTOASSERT(0, "unknown baudrate");
			   MO_ProtocolErrors++;
	}

	if (baudrate && AppPrefs.baudrateChanges)
		Serial_SetBaudRate (baudrate);

	return(0);
}


/* Escape Function
 * 		254 53
 * 		(Esc 5)
 * Read serial number
 */
static Int32 
escape_OutputSerialNo(UInt8 *ignored)
{
	ASSERT(ignored[0]==53, "wrong func");
	
	sendSerialNum();

	return(0);
}

static void
sendSerialNum(void)
{
	Serial_SendChar (usedEmuState.serialNum >> 8);
	Serial_SendChar (usedEmuState.serialNum & 0xFF);
}

/* Escape Function
 * 		254 52 <byte1> <byte2>
 * 		(Esc 4)
 * Set serial number
 */
static Int32 
escape_SetSerialNo(UInt8 *args) 
{
	ASSERT(args[0]==52, "wrong func");

	// Ignore if serial number already set...
	if (usedEmuState.serialSet == 0)
	{
		PROTOASSERT(usedEmuState.serialSet != 0, "Serial number already set");
		MO_ProtocolErrors++;
		return(0);
	}

	// Serial number set...
	usedEmuState.serialNum = (args[1] << 8) + args[2];
	usedEmuState.serialSet = 1; // stop it being changed again.

	sendSerialNum(); // The manual says that the serial is then outputed.

	return(0);
}



/* Escape Function
 * 		254 67
 * 		(Esc C)
 * Linewrap ON
 */
static Int32 
escape_SetLinewrapOn(UInt8 *ignored) 
{
	ASSERT(ignored[0]==67, "wrong func");

	usedEmuState.linewrap = true;
	return(0);
}


/* Escape Function
 * 		254 68
 * 		(Esc D)
 * Linewrap OFF
 */
static Int32 
escape_SetLinewrapOff(UInt8 *ignored) 
{
	ASSERT(ignored[0]==68, "wrong func");

	usedEmuState.linewrap = false;
	return(0);
}

			
/* Escape Function
 * 		254 81 
 * 		(Esc Q)
 * Autoscroll ON
 */
static Int32 
escape_SetAutoscrollOn(UInt8 *ignored)
{
	ASSERT(ignored[0]==81, "wrong func");

	usedEmuState.autoscroll = true;
	return(0);
}

/* Escape Function
 * 		254 82 
 * 		(Esc R)
 * Autoscroll OFF
 */
static Int32 
escape_SetAutoscrollOff(UInt8 *ignored)
{
	ASSERT(ignored[0]==82, "wrong func");

	usedEmuState.autoscroll = false;
	return(0);
}



/* Escape Function
 * 		254 65
 * 		(Esc A)
 * Autotransmit ON
 */
static Int32 
escape_SetAutotransmitOn(UInt8 *ignored) 
{
	ASSERT(ignored[0]==65, "wrong func");

	usedEmuState.autotransmit=true;
	keyBufLen=0;
	return(0);
}

/* Escape Function
 * 		254 79
 * 		(Esc O)
 * Autotransmit OFF
 */
static Int32 
escape_SetAutotransmitOff(UInt8 *ignored) 
{
	ASSERT(ignored[0]==79, "wrong func");

	usedEmuState.autotransmit=false;
	keyBufLen=0;
	return(0);
}

/* Escape Function
 * 		254 83
 * 		(Esc S)
 * Block (blinking) Cursor ON
 */
static Int32 
escape_SetBlockOn(UInt8 *ignored)
{
	ASSERT(ignored[0]==83, "wrong func");

	notUsedEmuState.block = true;
	return(0);
}

/* Escape Function
 * 		254 84
 * 		(Esc T)
 * Block (blinking) Cursor OFF
 */
static Int32 
escape_SetBlockOff(UInt8 *ignored)
{
	ASSERT(ignored[0]==84, "wrong func");

	notUsedEmuState.block = false;
	return(0);
}

/* Escape Function
 * 		254 74
 * 		(Esc J)
 * Underline Cursor ON
 */
static Int32 
escape_SetUnderlineOn(UInt8 *ignored) 
{
	ASSERT(ignored[0]==74, "wrong func");

	//ASSERT(0, "Underline not implemented");

	notUsedEmuState.underline = true;
	return(0);
}

/* Escape Function
 * 		254 75
 * 		(Esc K)
 * Underline Cursor OFF
 */
static Int32
escape_SetUnderlineOff(UInt8 *ignored) 
{
	ASSERT(ignored[0]==75, "wrong func");

	notUsedEmuState.underline = false;
	return(0);
}


/* Escape Function
 * 		254 126 <mode>
 * 		(Esc ~)
 * Autorepeat ON
 * <mode>: 0 == 200 ms typematic,
 *         1 == key down/key up codes sent.
 */
static Int32 
escape_SetAutorepeatOn(UInt8 *args)
{
	UInt8 set = args[1];

	ASSERT(args[0]==126, "wrong func");

	if ((set==0) || (set==1))
	{
		notUsedEmuState.autorepeatMode = set;
		notUsedEmuState.autorepeat = 1;
	}
	else
	{
		PROTOASSERT((set==0) || (set==1), "Autorepeat: bad arg?");
		MO_ProtocolErrors++;
	}

	return (0);
}

/* Escape Function
 * 		254 96
 * 		(Esc `)
 * Autorepeat OFF
 */
static Int32 
escape_SetAutorepeatOff(UInt8 *ignored)
{
	ASSERT(ignored[0]==96, "wrong func");

	notUsedEmuState.autorepeat = 0;
	return(0);
}


/* Escape Function
 * 		254 72
 * 		(Esc H)
 * move cursor to home
 */
static Int32 
escape_Home (UInt8 *ignored) 
{
	ASSERT(ignored[0]==72, "wrong func");

	mustAutoscroll = 0;
	internalEmuState.xPos = 0;
	internalEmuState.yPos = 0;

	return 0;
}


/* Escape Function
 * 		254 66 <mins>
 * 		(Esc B)
 * Backlight: turn ON backlight
 * for <mins> minutes. 0 == Always.
 */
static Int32 
escape_SetBacklightOn(UInt8 *args)
{
	ASSERT(args[0]==66, "wrong func");

	//EvtEnqueueKey(vchrBacklight, 0, commandKeyMask); // toggles...

	notUsedEmuState.backlightMins = args[1];
	usedEmuState.backlight = true; 
	updateBacklight();
	return (0);
}

/* Escape Function
 * 		254 70
 * 		(Esc F)
 * Backlight: turn OFF backlight.
 *
 * CJL: This was coded to toggle - not sure why...
 */
static Int32 
escape_SetBacklightOff(UInt8 *ignored)
{
	ASSERT(ignored[0]==70, "wrong func");

	usedEmuState.backlight = false;
	updateBacklight();
	return (0);
}

/* Escape Function
 * 		254 58 <full> <empty>
 * 		(Esc :)
 * Flow control ON (RS232C only)
 */
static Int32 
escape_SetFlowControlOn(UInt8 *args)
{
	ASSERT(args[0]==58, "wrong func");

	//ASSERT(0, "Flow Control not implemented");

	notUsedEmuState.flowControlFull = args[1];
	notUsedEmuState.flowControlEmpty = args[2];
	notUsedEmuState.flowControlMode = 1;
	return(0);
}

/* Escape Function
 * 		254 59
 * 		(Esc ;)
 * Exit flow control mode
 */
static Int32 
escape_SetFlowControlOff(UInt8 *ignored)
{
	ASSERT(ignored[0]==59, "wrong func");

	notUsedEmuState.flowControlMode = 0;
	return(0);
}

/* Escape Function
 * 		254 71 <col> <row>
 * 		(Esc G)
 * move to position (x,y)
 */
static Int32 
escape_MoveToXY(UInt8 *args)
{
	UInt32 x = args[1];
	UInt32 y = args[2];
	ASSERT(args[0]==71, "wrong func");

	if ((x<=internalEmuState.width) && (x>0)
			&& (y<=internalEmuState.height) && (y>0))
	{
		mustAutoscroll = 0;
		goToXY(x-1, y-1);
	}
	else
	{
		PROTOASSERT2(0, "MoveToXY(%u,%u): out of range", x, y);
		MO_ProtocolErrors++;
	}

	return 0;
}


/* Escape Function
 * 		254 51 <address>
 * 		(Esc 3)
 * Set the I2C Address
 */
static Int32 
escape_SetI2CAddress(UInt8 *args)
{
	UInt8 addr = args[1];
	ASSERT(args[0]==51, "wrong func");

	// address must be even.
	if (addr%2)
	{
		notUsedEmuState.addressI2C = addr;
	} else {
		PROTOASSERT(!(addr%2), "IC2 address not even");
		MO_ProtocolErrors++;
	}
	return(0);
}


/* Escape Function
 * 		254 80 <contrast>
 * 		(Esc P)
 * 	Set Contrast
*
* 		254 145 <contrast>
 * 	Set Contrast and remember setting
 */
static Int32 
escape_SetContrast(UInt8 *args)
{
	ASSERT((args[0]==80)||(args[0]==145), "wrong func");

	notUsedEmuState.contrast = args[1];

	return (0);
}


/* Escape Function
 * 		254 87 <gpo#>  
 * 		(Esc W)
 * Turn ON gpo <gpo#>
 * [We just show the gpo status on screen]
 * <gpo#> is 1..internalEmuState.numOutputs [currently 6]
 */
static Int32 
escape_SetOutputOn(UInt8 *args)
{
	UInt8 out = args[1];

	ASSERT(args[0]==87, "wrong func");

	if ((out>0) && (out <= internalEmuState.numOutputs))
	{
		Screen_setOutput(out, true);
	} else {
		PROTOASSERT(out<=internalEmuState.numOutputs, "Output out of range");
		MO_ProtocolErrors++;
	}

	return(0);
}

/* Escape Function
 * 		254 86 <gpo#>  
 * 		(Esc V)
 * Turn OFF gpo <gpo#>
 * [We just show the gpo status on screen]
 * <gpo#> is 1..internalEmuState.numOutputs [currently 6]
 */
static Int32 
escape_SetOutputOff(UInt8 *args)
{
	UInt8 out = args[1];

	ASSERT(args[0]==86, "wrong func");

	if ((out> 0) && (out <= internalEmuState.numOutputs))
	{
		Screen_setOutput(out, false);
	} 
	else 
	{
		/* This assert fires when using LCDd with type 'lcd' rather than
		 * 'lkd' - they assume 'lcd' has only 1 output and send only
		 * 254 86 (without a gpo#) - this causes protocol errors.
		 */
		PROTOASSERT(out<=internalEmuState.numOutputs, "Output out of range");
		MO_ProtocolErrors++;
	}

	return(0);
}


/* Escape Function
 * 		254 64 <char>*40   (or 80)
 * 		(Esc @)
 * Load startup screen and save it...
 *
 * Currently assuming 80 chars...
 */
static Int32 
escape_SetStartupScreen(UInt8 *args)
{
	ASSERT(args[0]==64, "wrong func");

	ASSERT(internalEmuState.height*internalEmuState.width == 80, "lcd h*w!=80");
	//IGNORED
	//MemMove(AppPrefs.StartupScreen, args, 80);
	return(0);
}


/* Escape Function
 * 		254 115
 * 		(Esc s)
 * Initialize for thin vertical bar graphs
 */
static Int32 
escape_InitVBarNarrow(UInt8 *ignored)
{
	ASSERT(ignored[0]==115, "wrong func");

	// where does the narrow bar go exactly?
	Screen_DefineChar(0, (UInt8*)"\x00\x00\x00\x00\x00\x00\x00\x06");// layer 1 
	Screen_DefineChar(1, (UInt8*)"\x00\x00\x00\x00\x00\x00\x06\x06");// layer 2
	Screen_DefineChar(2, (UInt8*)"\x00\x00\x00\x00\x00\x06\x06\x06");// layer 3
	Screen_DefineChar(3, (UInt8*)"\x00\x00\x00\x00\x06\x06\x06\x06");// layer 4
	Screen_DefineChar(4, (UInt8*)"\x00\x00\x00\x06\x06\x06\x06\x06");// layer 5
	Screen_DefineChar(5, (UInt8*)"\x00\x00\x06\x06\x06\x06\x06\x06");// layer 6 
	Screen_DefineChar(6, (UInt8*)"\x00\x06\x06\x06\x06\x06\x06\x06");// layer 7
	Screen_DefineChar(7, (UInt8*)"\x06\x06\x06\x06\x06\x06\x06\x06");// layer 8
	return(0);
}

/* Escape Function
 * 		254 118 
 * 		(Esc v)
 * Initialize for wide vertical bar graphs
 */
static Int32 
escape_InitVBarWide(UInt8 *args)
{
	ASSERT(args[0]==118, "wrong func");

	Screen_DefineChar(0, (UInt8*)"\x00\x00\x00\x00\x00\x00\x00\x1F"); // layer 1
	Screen_DefineChar(1, (UInt8*)"\x00\x00\x00\x00\x00\x00\x1F\x1F"); // layer 2
	Screen_DefineChar(2, (UInt8*)"\x00\x00\x00\x00\x00\x1F\x1F\x1F"); // layer 3
	Screen_DefineChar(3, (UInt8*)"\x00\x00\x00\x00\x1F\x1F\x1F\x1F"); // layer 4
	Screen_DefineChar(4, (UInt8*)"\x00\x00\x00\x1F\x1F\x1F\x1F\x1F"); // layer 5
	Screen_DefineChar(5, (UInt8*)"\x00\x00\x1F\x1F\x1F\x1F\x1F\x1F"); // layer 6
	Screen_DefineChar(6, (UInt8*)"\x00\x1F\x1F\x1F\x1F\x1F\x1F\x1F"); // layer 7
	Screen_DefineChar(7, (UInt8*)"\x1F\x1F\x1F\x1F\x1F\x1F\x1F\x1F"); // layer 8

	return(0);
}

				
/* Escape Function
 * 		254 104
 * 		(Esc h)
 * Initialize for horizontal bar graphs
 */
static Int32 
escape_InitHBar(UInt8 *ignored)
{
	ASSERT(ignored[0]==104, "wrong func");

	Screen_DefineChar(0, (UInt8*)"\x10\x10\x10\x10\x10\x10\x10\x00");//right(1x)
	Screen_DefineChar(1, (UInt8*)"\x18\x18\x18\x18\x18\x18\x18\x00");//right(2x)
	Screen_DefineChar(2, (UInt8*)"\x1C\x1C\x1C\x1C\x1C\x1C\x1C\x00");//right(3x)
	Screen_DefineChar(3, (UInt8*)"\x1E\x1E\x1E\x1E\x1E\x1E\x1E\x00");//right(4x)

	Screen_DefineChar(4, (UInt8*)"\x01\x01\x01\x01\x01\x01\x01\x00");//left(1x)
	Screen_DefineChar(5, (UInt8*)"\x03\x03\x03\x03\x03\x03\x03\x00");//left(2x)
	Screen_DefineChar(6, (UInt8*)"\x07\x07\x07\x07\x07\x07\x07\x00");//left(3x)
	Screen_DefineChar(7, (UInt8*)"\x0F\x0F\x0F\x0F\x0F\x0F\x0F\x00");//left(4x)

	return (0);
}


/* Escape Function
 * 		254 110
 * 		(Esc n)
 * Init Large digits
 */
static Int32 
escape_InitBigNum(UInt8 *ignored)
{
	ASSERT(ignored[0]==110, "wrong func");

	// top ---> bottom...
	Screen_DefineChar(0, (UInt8*)"\x1F\x1F\x1F\x1F\x00\x00\x00\x00");//top half
	Screen_DefineChar(1, (UInt8*)"\x00\x00\x00\x1F\x1F\x1F\x1F\x00");//bot half

	Screen_DefineChar(2, (UInt8*)"\x1C\x1C\x1C\x1C\x1C\x1C\x1C\x00");//left half
	Screen_DefineChar(3, (UInt8*)"\x03\x03\x03\x03\x03\x03\x03\x00");//righthalf

	Screen_DefineChar(4, (UInt8*)"\x1F\x1F\x1E\x1C\x1C\x18\x10\x00");//tl corner
	Screen_DefineChar(5, (UInt8*)"\x01\x03\x03\x07\x0F\x0F\x1F\x00");//br corner

	Screen_DefineChar(6, (UInt8*)"\x1F\x0F\x0F\x07\x03\x03\x01\x00");//tr corner
	Screen_DefineChar(7, (UInt8*)"\x10\x18\x18\x1C\x1E\x1E\x1F\x00");//bl corner

	return (0);
}



/* Escape Function
 * 		254 35 <col> <digit>
 * 		(Esc #)
 * Big Numbers
 *
 */
static Int32 
escape_displayBigNum(UInt8 *args)
{
	UInt16 x, digit;

	ASSERT(args[0]==35, "wrong func");

	x = args[1];
	digit = args[2];
	if ((digit<=9) && (x<=internalEmuState.width)
			&& (x>0))
	{
		displayBigNum(digit, x-1);
	}
	else
	{
		PROTOASSERT(0, "big num proto error");
		MO_ProtocolErrors++;
	}

	return (0);
}



/* Escape Function
 * 		254 61 <col> <length>
 * 		(Esc =)
 * Draw vertical bar graph
 * Draws a vertical bar graph at column [col] of length [length]. 
 * Length is measured in pixels (0x00 to 0x14).
 * User must first use the 'v' or 's' command to initialize characters.
 */
static Int32 
escape_drawVBar(UInt8 *args)
{
	UInt16 col, barlen;

	ASSERT(args[0]==61, "wrong func");

	col = args[1];
	barlen = args[2];

	if ((barlen<=(internalEmuState.height*FONTCELL_HEIGHT)) &&
		   	(col<=internalEmuState.width) && (col>0))
	{
		drawVBar(col-1, barlen);
	}
	else
	{
		PROTOASSERT2(0, "drawVBar proto error (col=%u, barlen=%u)", col, barlen);
		MO_ProtocolErrors++;
	}

	return (0);
}

/* Expects 0-based coords.
 */
static void
drawVBar(UInt16 col, UInt16 barlen) 
{
    UInt16 y, i, fullbars, partialbar;

    internalEmuState.xPos = col;
    if (internalEmuState.xPos >= internalEmuState.width)
        internalEmuState.xPos = internalEmuState.width - 1;

    fullbars = 0;
    partialbar = 0;

    internalEmuState.yPos = internalEmuState.height - 1;

    if (barlen == 0) 
    {
        for (y = 0; y < internalEmuState.height; y++)
        {
            Screen_PutChar(' ', internalEmuState.xPos, y);
        }
    } else {
        fullbars = (UInt16) (barlen / FONTCELL_HEIGHT);
        partialbar = (UInt16) (barlen % FONTCELL_HEIGHT);

        // Draw full bars, if any...

        if (fullbars > 0)
        {

            for (i = 0; i < fullbars; i++)
            {

                // don't draw off screen...
                if ((internalEmuState.yPos - i) >= 0)
                {
					// custom char 7 has previously been set to be a full bar.
                    Screen_PutChar(0x07, internalEmuState.xPos, internalEmuState.yPos-i);
                }
            }
        }
    }

    // Draw partial bars, if any....

    if (partialbar != 0)
    {
        // don't draw off-screen
        if (fullbars < internalEmuState.height)
        {
            Screen_PutChar(partialbar-1, internalEmuState.xPos, internalEmuState.yPos-fullbars);
        }
    }

    //modified = 1;
}


/* Escape Function
 * 		254 124 <c> <r> <d> <length>
 * 		(Esc |)
 * Horizontal bar graphs
 * Draws a horizontal bar graph starting at column [c] on row [r] with
 * direction [d] 0 is right, 1 is left) of length [length]. 
 * Length is measured in pixels (0x00 to 0x64 if starting in column 1).
 * User must first use the 'h' command to initialize characters.
 */
static Int32 
escape_DrawHBar(UInt8 *args)
{
	UInt16 x,y, barlen, dir;

	ASSERT(args[0]==124, "wrong func");

	x = args[1];
	y = args[2];
	dir = args[3];
	barlen = args[4];

	if ((dir<=1) && (barlen<=(internalEmuState.width*FONTCELL_WIDTH))
			&& (x>0) && (y>0) 
			&& (x<=internalEmuState.width) 
			&& (y<=internalEmuState.height))
	{
		goToXY(x-1, y-1);
		drawHorizontalBar(dir, barlen);
	}
   	else
	{
		PROTOASSERT(0, "DrawHBar proto error");
		MO_ProtocolErrors++;
	}

	return (0);
}

static void
drawHorizontalBar(UInt16 dir, UInt16 barlen)
{
    UInt16 i, fullbars, partialbar;

    //  dir: 00 - left to right
    //  dir: 01 - right to left

    dir &= 0x01;

    if (barlen == 0) 
        return;

    fullbars = (UInt16) (barlen / FONTCELL_WIDTH);
    partialbar = (UInt16) (barlen % FONTCELL_WIDTH);

    // Draw full bars, if any...

    if (fullbars > 0)
    {
        for (i = 0; i < fullbars; i++)
        {
            if (dir == 0)
            {
                // don't draw off screen...
                if ((internalEmuState.xPos + i) < internalEmuState.width)
                {
                    Screen_PutChar(FULL_BLOCK, internalEmuState.xPos+i, internalEmuState.yPos);
                }
            } else {
                // don't draw off screen...
                if ((internalEmuState.xPos - i) >= 0)
                {
                    Screen_PutChar(FULL_BLOCK, internalEmuState.xPos-i, internalEmuState.yPos);
                }
            }
        }
    }

    // Draw partial bars, if any....

    if (partialbar != 0)
    {
        if (dir == 0)
        {
            // don't draw off-screen
            if ((internalEmuState.xPos + fullbars) < internalEmuState.width)
            {
                Screen_PutChar(partialbar-1, internalEmuState.xPos+fullbars, internalEmuState.yPos);
            }
        } else {
            // don't draw off-screen
            if ((internalEmuState.xPos - fullbars) > 0)
            {
                Screen_PutChar(partialbar-1+4, internalEmuState.xPos-fullbars, internalEmuState.yPos);
            }
        }
    }
}


static void
debug_startOnNewline (void) 
{
	doCarriageReturn();
	doLineFeed();
}

static void 
debug_outputString (Char *str) 
{
	while (*str) 
	{
		putOutChar(*(str++));
	}
}

static Char debug_hexDigits[]="0123456789ABCDEF";

static void
debug_outputHex(UInt8 byte) 
{
	if (internalEmuState.xPos + 2 >= internalEmuState.width) 
	{
		debug_startOnNewline();
	}

	putOutChar(debug_hexDigits[byte>>4]);
	putOutChar(debug_hexDigits[byte & 0x0F]);
}



static void
debug_outputChar (UInt8 chr) 
{
	if (chr < NUM_SPECIAL_CHARS)
	{
		debug_outputString("<u");
		putOutChar('0' + chr);
		putOutChar('>');
		return;
	}

	switch(chr) {
		case '\n':
			debug_outputString("<LF>");
			break;
		case 0x8:
			debug_outputString("<BS>");
			break;
		case 0xC:
			debug_outputString("<ClrScrn>");
			break;
		case '\r':
			debug_outputString("<CR>");
			break;
		case ' ':
			debug_outputString("<SP>");
			break;
		default:
			//debug_outputHex(chr);
			putOutChar(chr);
			break;
	}

#if DEBUG
	Screen_UpdateLcd();
#endif
}


static void 
debug_hexOutput (UInt8 *buf, UInt32 size) 
{
	while (size)
	{
		UInt8 chr = *buf;

		if (chr == CMD_CHAR)
			debug_startOnNewline();

		debug_outputHex(chr);

		size--;
		buf++;
	}

#if DEBUG
	Screen_UpdateLcd();
#endif
}

static void
debug_displayFunc(UInt8 *buf)
{
	UInt8 command = *buf;
	UInt32 commandLen = ARGCOUNT(command);

	if (internalEmuState.xPos!=0)
		debug_startOnNewline();
	debug_outputString(CMDNAME(command));
	putOutChar(':');
	for (UInt32 i=1; i<=commandLen; i++)
	{
		debug_outputHex(buf[i]);
		if (i<commandLen)
			putOutChar(',');
	}
	debug_startOnNewline();

#if DEBUG
	Screen_UpdateLcd();
#endif
}

static void
debug_displayBadFunc(UInt8 cmd)
{
	debug_outputString("BadCmd ");
	debug_outputHex(cmd);
	debug_startOnNewline();
#if DEBUG
	Screen_UpdateLcd();
#endif
}

/*
 * Process received data...
 *
 */

void
MO_ProcessInput(UInt8 *chr, UInt32 size) 
{
	static UInt8 leftoverBuf[MAX_CMD_LENGTH]; // See Note 2 in NOTES
	static UInt32 leftover=0;

	ASSERT(size!=0, "size is 0");
	ASSERT(chr, "chr is NULL");

	if (showHex) 
	{
		// Just display the received data in hex.
		debug_hexOutput(chr, size);
		return;
	}

	
	while (size)
	{
		UInt8 firstByte;

		if (leftover)
			firstByte=*leftoverBuf;
		else
			firstByte=*chr;

		if (firstByte == CMD_CHAR)
		{   /* it's command */

			UInt8   command;
			UInt8   commandLen;
			cmd_func commandFunc;

			/* we need at least 2 bytes to check what the command is! */
			if ((size+leftover)<2)
			{
				/* we need more data */
				ASSERT(size+leftover <= MAX_CMD_LENGTH, "data too large");
				MemMove(leftoverBuf+leftover, chr, size);
				leftover+=size;
				return;
			}

			if (leftover>1)
				command = leftoverBuf[1];
			else
				command = chr[1-leftover];

			// Check that the command code is a valid one
			if ((command > MAX_CMD_CODE) || (command < MIN_CMD_CODE))
			{
				// bad command - out of range
				PROTOASSERT1(0, "bad command: %d", command);
				MO_ProtocolErrors++;
				if ((MO_ProtocolErrors>100) && (AppPrefs.warnProtocolErrors))
				{
					static MInt8 alreadyWarned=0;
					if (!alreadyWarned)
					{
						alreadyWarned=1;
						FrmAlert(ErrProtocolErrorsAlert);
					}
				}

				if (showCmds)
					debug_displayBadFunc(command);

				// leftover has to be either 0 or 1; it can't be
				// 2 or higher because otherwise it would have already
				// been here and been discarded!
				ASSERT(leftover<=1, "leftover assumption wrong");

				size-=(2-leftover);
				chr+=(2-leftover);
				leftover=0;
			}
			else if (!(commandFunc=CMDFUNC(command)))
			{
				// bad command - not such function
				PROTOASSERT(CMDFUNC(command), "bad command - no function");
				MO_ProtocolErrors++;
				if (showCmds)
					debug_displayBadFunc(command);

				// leftover has to be either 0 or 1; it can't be
				// 2 or higher because otherwise it would have already
				// been here and been discarded!
				ASSERT(leftover<=1, "leftover assumption wrong");

				size-=(2-leftover);
				chr+=(2-leftover);
				leftover=0;
			}
			else
			{   /* it's a valid command */
				commandLen = ARGCOUNT(command) + 2; /*2 to inc esc & cmdcode */

				if ((size+leftover) < commandLen)
				{
					/* we need more data */
					ASSERT(size+leftover <= MAX_CMD_LENGTH, "data too large");
					MemMove(leftoverBuf+leftover, chr, size);
					leftover+=size;
					return;
				}


				if (!leftover)
				{
					// The command is complete, now process it.
					if (!showCmds)
					{
						commandFunc(chr+1); //+1 as the escape code is of no interest.
					}
					else
					{
						debug_displayFunc(chr+1);
					}
					size -= commandLen;
					chr += commandLen;
				}
				else
				{
					ASSERT(commandLen <= MAX_CMD_LENGTH, "data too large");
					MemMove(leftoverBuf+leftover, chr, commandLen-leftover);
					if (!showCmds)
					{
						commandFunc(leftoverBuf+1);
					}
					else
					{
						debug_displayFunc(leftoverBuf+1);
					}

					size -= (commandLen-leftover);
					chr += (commandLen-leftover);
					leftover = 0;
				}

			}
		} 
		else
		{   // Must be a regular character: print to the LCD.
			if (!showCmds)
			{
				switch (firstByte) 
				{
					case '\n':
						doLineFeed();
						break;

					case '\r':
						doCarriageReturn();
						break;

					case 0x8:
						retreatPos();
						// Manual says:	Moves the cursor one position to the
						// left and clears that position.
						Screen_PutChar(' ', internalEmuState.xPos,
							   				internalEmuState.yPos);
						break;

					case 0xC:
						MO_ClearScreen();
						break;

					default:
						putOutChar(firstByte);
						break;
				}
			}
			else
			{
				debug_outputChar(firstByte);
			}

			ASSERT(leftover==0, "leftover assumption wrong");
			size--;
			chr++;
		}

	}

	/* If we needed more data then we would have returned already.
	 * So getting here means all the data has been used.
	 */
	ASSERT((size==0) && (leftover==0), "size assumption wrong");
	return;
}


