/******************************************************************************
 *
 *  PalmOrb - Emulate a Matrix Orbital LCD Display with Keypad
 *  Copyright (C) 2004  Chris Lansley  <palmorb@lansley.me.uk>
 *  Copyright (C) 2001  David Douthitt 
 *  Copyright (C) 2000  J Robert Ray
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  $Source: /cvsroot/palmorb/palmorb/MatrixOrbital.h,v $
 *  $Revision: 1.4 $ $Date: 2004/09/26 16:18:48 $
 *
 ******************************************************************************/

#ifndef __MATRIXORBITAL_H__
#define __MATRIXORBITAL_H__

#include "segments.h"

/*
 * NB: This structure is STORED in user preferences
 *     any changes need to be handled and the preferences
 *     version number needs changing.
 */
typedef struct UsedEmulationState
{ /* UsedEmulationState is settable by using MO_SetState.
   *
   * These can be set directly by sending commands via the serial port.
   * Some can be read by sending commands via the serial port.
   */
    UInt8  autoscroll;      // automatic scroll
    UInt8  linewrap;        // auto line wrap
    UInt8  autotransmit;    // auto send keys
    UInt8  serialSet;		// serial number set?
    UInt16 serialNum;		// Serial number of device
    UInt8  backlight;		// backlight state

	UInt8  padding[16];		// padding to aid later adding to this structure.
};

typedef struct NotUsedEmulationState
{ /* NotUsedEmulationState is NOT settable except internally in
   * MatrixOrbital.cpp. You can get the settings by using MO_GetState.
   *
   * These can be set directly by sending commands via the serial port.
   * Some can be read by sending commands via the serial port.
   *
   * These settings have NO effect and are simply noted and given back
   * if asked.
   */

	UInt8  flowControlFull;		// IGNORED
	UInt8  flowControlEmpty;	// IGNORED
	UInt8  flowControlMode;		// IGNORED
    UInt8  backlightMins;		// backlight state  - IGNORED
    UInt8  underline;			// underline cursor - IGNORED
    UInt8  block;				// block cursor     - IGNORED
    UInt8  autorepeat;			// keypad autorepeat - IGNORED
    UInt8  autorepeatMode;		// keypad autorepeat - IGNORED
    UInt8  contrast;			// contrast level    - IGNORED
    UInt8  debounceTime;		// keypad debounce time - IGNORED
    UInt8  addressI2C;			// I2C address - IGNORED
};

typedef struct InternalEmulationState
{ /* InternalEmulationState is NOT settable except internally in
   * MatrixOrbital.cpp. You can get the settings by using MO_GetState.
   *
   * Internal State that may be of interest to other parts of the code.
   */
	UInt32 xPos;  // Can also be set via serial port using GotoXY cmd
	UInt32 yPos;  // Can also be set via serial port using GotoXY cmd
	UInt32 numOutputs;
	UInt32 numKeys;
	UInt32 width;
	UInt32 height;
};


/* Exported functions */
extern Err MO_Init(UsedEmulationState *defaults) SEG2;
extern void MO_Fini(void) SEG2;
extern void MO_ProcessInput(UInt8 *chr, UInt32 size); // feed serial data in
														
extern void MO_ClearOutputs(void); // for palmorb.cpp to reset the outputs
extern void MO_ClearScreen(void); // for palmorb.cpp to clear LCD screen
extern void MO_SendKey (Char key); 
extern void MO_enterParsedMode(void);
extern void MO_enterHexMode(void);
extern void MO_returnToNormalMode(void);
extern void MO_SetState(UsedEmulationState *state);
extern void MO_GetState(UsedEmulationState *used,
	   					NotUsedEmulationState *notused,
						InternalEmulationState *internal);
extern void MO_PrefsChanged(void);

/* Exported variables */
extern Boolean MO_ShowHex; // set in palmorb.cpp - show hex rather than lcd
extern Boolean MO_ShowCmds;// set in palmorb.cpp - show cmds rather than lcd
extern UInt32 MO_ProtocolErrors; // for palmorb.cpp to display stats


// Maximum allowable outputs...
#define MAX_OUTPUTS 8
#define NUM_SPECIAL_CHARS 8

// 5x8 character cells
#define FONTCELL_HEIGHT 8
#define FONTCELL_WIDTH 5

#endif //  __MATRIXORBITAL_H__
