/******************************************************************************
 *
 *  PalmOrb - Emulate a Matrix Orbital LCD Display with Keypad
 *  Copyright (C) 2004  Chris Lansley  <palmorb@lansley.me.uk>
 *  Copyright (C) 2001  David Douthitt 
 *  Copyright (C) 2000  J Robert Ray
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  $Source: /cvsroot/palmorb/palmorb/palmorb.cpp,v $
 *  $Revision: 1.19 $ $Date: 2004/10/22 12:43:53 $
 *
 ******************************************************************************/

        /*************************************************/
        /* This file contains mostly the GUI code.       */
        /* Also the main event loop and main entry point */
        /*************************************************/

/**
 * Orginally we were emulating a LK202-25 (a 20x2 lcd) but with a 20x4 screen.
 * I think it's better to emulate a LK204-25 which has a 20x4 screen.
 */


#include <PalmOS.h>

#include "resources.h"
#include "MatrixOrbital.h"
#include "palmorb.h"
#include "screen.h"
#include "serial.h"
#include "debug.h"
#include "segments.h"
#include "ui.h"

/* Exported functions */
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags);
void PalmOrb_SysInit(void) SEG2;
void Palmorb_SetSerialDefaults(void) SEG2;
void Palmorb_SetEmuDefaults(UsedEmulationState *defaults) SEG2;

/* Exported Variables */
ApplicationPreferences AppPrefs;	// These are remembered.

/* File Global variables */
static Boolean lowBattery = false;
#if DEBUG_CAPTURE
static UInt8 *debug_CapBuf;
static Boolean debug_CapWrapped;
static UInt8 *debug_CapIns;
static UInt8 *debug_CapEnd;
#define CAPTURE_SIZE (30L * 1024)
#endif // DEBUG_CAPTURE
static Boolean newSysBatteryInfo = false;

/* Local functions */
static Int16 loadAppPrefs(void) SEG2;
static Err startApp(void) SEG2;
static void stopApp(void) SEG2;
static void processKey(UInt16 ch);
static void specialPreProcessing(EventPtr event);
static void specialPostProcessing(EventPtr event);
static void readSerial(void);
static void eventLoop(void);
static Err romVersionCompatible(UInt16 launchFlags);


static void
powerManagement(Boolean saidToBeLow)
{
	Boolean PluggedIn;
	UInt16 warnThreshold;
	UInt16 curLevel;
	UInt8 percent = 0;


	// Only check the battery level and reset the timer if we haven't been
	// here in the last 10 seconds! [Unless the system has told us it's low]
	if (!saidToBeLow)
	{
		static UInt32 lasttime = 0;
		UInt32 now = TimGetTicks();
		static UInt16 tick_wait = SysTicksPerSecond() * 10;
		if ((lasttime + tick_wait >= now) && (now >= lasttime))
			return;
		lasttime = now;
	}

	FormType *frm = FrmGetFormPtr(MainFormID);
	//ASSERT(frm, "Main not loaded");
	if (!frm)
		return;


	// lowBattery may not be accurate, either because
	// we just started (no warnings yet) or because a new
	// battery was put in.  So we do our own checks...

	if (newSysBatteryInfo)
		curLevel = SysBatteryInfo(false, &warnThreshold, NULL, NULL, NULL,
								  &PluggedIn, &percent);
	else
		curLevel = SysBatteryInfoV20(false, &warnThreshold, NULL, NULL, NULL,
									 &PluggedIn);

	if (PluggedIn == true)
	{
		if (lowBattery == true)
		{
			FrmAlert(PluggedInAlert);
			lowBattery = false;	// no longer dependent on battery...
			UI_showPowerStatus(frm, lowBattery);
		}
	}
	else
	{	// !PluggedIn

		if (curLevel > warnThreshold)
		{	// good battery
			lowBattery = false;
			UI_showPowerStatus(frm, lowBattery);
		}
		else
		{	// low battery
			if (lowBattery == false)
			{
				Char buf[100];

				if (newSysBatteryInfo)
					StrPrintF(buf, "%u%%", percent);
				else
					StrPrintF(buf, "%u.%u volts", curLevel/100, curLevel%100);

				FrmCustomAlert(MyLowBatteryAlert, buf, " ", " ");
				lowBattery = true;
				UI_showPowerStatus(frm, lowBattery);
			}
		}
	}

	// this assert fired a lot - seems the OS tell us the battery is 
	// low even if it isn't...
	//ASSERT((!saidToBeLow) || lowBattery, "said to be low, but doesn't seem so");

	// Reset auto OFF: Keep from turning off... only
	// if battery is not low, and only if not to turn off automatically...
	if (lowBattery == false)
		EvtResetAutoOffTimer();
}

void
Palmorb_SetEmuDefaults(UsedEmulationState *defaults)	//SEG2
{
	defaults->autoscroll = 0;	// automatic scroll
	defaults->linewrap = 1;		// auto line wrap - real device has default of 0
	defaults->autotransmit = 1;	//
	defaults->serialSet = 0;	// serial number not yet set
	defaults->serialNum = 0;	// serial number (unset)
	defaults->backlight = 0;	// 
}

void
Palmorb_SetSerialDefaults(void)	// SEG2
{
	AppPrefs.serialBaud = 19200;
	AppPrefs.serialParity = SERIAL_PARITY_NONE;
	AppPrefs.serialStopBits = 1;
	AppPrefs.serialDataBits = 8;
	AppPrefs.serialFlowCtl = 1;
	AppPrefs.serialErrDetails=0;
	AppPrefs.serialBigBuf=0;
	AppPrefs.serialDev = SERIAL_DEV_AUTO;
}

static Int16
loadAppPrefs(void)	//SEG2
{
	UInt16 prefsSize;
	Int16 prefsVersion = noPreferenceFound;

	prefsSize = sizeof(AppPrefs);
	prefsVersion = PrefGetAppPreferences((UInt32)'POrb', 1,
										&AppPrefs, &prefsSize, true);

	// If there is a size mismatch with the current version - then
	// ignore them.
	if (prefsVersion == CurrentPrefVersion)
	{
		//ASSERT(prefsSize == sizeof(AppPrefs), "pref size mismatch");
		if (prefsSize != sizeof(AppPrefs))
			prefsVersion = noPreferenceFound;
	}


	if (prefsVersion > CurrentPrefVersion)
		prefsVersion = noPreferenceFound;	// found newer prefs: overwrite them

	switch (prefsVersion)
	{
		case CurrentPrefVersion:
			break;

	    case 8:
			AppPrefs.serialErrDetails=0;
			AppPrefs.serialBigBuf=0;
            /* fall though */
		case 9:
			AppPrefs.serialDev = SERIAL_DEV_AUTO;
            /* fall though */

		case 10:
			AppPrefs.sendGraffiti = false;
            /* fall though */

		case 11:
			AppPrefs.sendKeyUpDown = false;
			break;

		default:
			/* too old or newer version found - reset them */
			Palmorb_SetEmuDefaults(&AppPrefs.defaults);
			Palmorb_SetSerialDefaults();

			AppPrefs.colorMode = COLORMODE_GREEN;
			AppPrefs.showKeyboard = 1;
			AppPrefs.showOutputs = 1;
			AppPrefs.showCustomChars = 1;
			AppPrefs.backlight = 0;
			AppPrefs.output1Beep = 0;
			AppPrefs.refreshRate = 20;
			AppPrefs.baudrateChanges = 0;
			AppPrefs.correctSlash = 0;
			AppPrefs.padCharCells = 1;
			AppPrefs.warnSerialErrors = 1;
			AppPrefs.warnProtocolErrors = 1;
			AppPrefs.seenFeedBack = 0;
			AppPrefs.sendKeyUpDown = false;
			break;
	}

	return CurrentPrefVersion;
}


void
PalmOrb_SysInit(void)	//SEG2
{
	loadAppPrefs();
	MO_SetState(&AppPrefs.defaults);

	//Screen_SetColorMode(AppPrefs.colorMode);

	//Screen_UpdateLcd();
}


/*
 * Start Palm Application...
 *
 */

static Err
startApp(void)	//SEG2
{
	Err error = 0;

#if DEBUG
	DebugInit();
#endif

	loadAppPrefs();

	// Screen_Init must be called before MO_Init this is because
	// MO_Init will be setting the screen size during MO_Init
	error = Screen_Init();
	if (error != errNone)
		return (error);

	error = MO_Init(&AppPrefs.defaults);
	if (error != errNone)
		return (error);

	error = Serial_Init();
	if (error != errNone)
		return (error);

	// Check if the new SysBatteryInfo is available;
	UInt32 romVersion;
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

	if (romVersion >= sysMakeROMVersion(3, 0, 0, sysROMStageRelease, 0))
		newSysBatteryInfo = true;

	lowBattery = false;	// at least until we check (later)...


#if DEBUG_CAPTURE
	debug_CapBuf = (UInt8 *) MemPtrNew(CAPTURE_SIZE);
	if (!debug_CapBuf)
	{
		SysFatalAlert("Can't allocate capture buffer");
		// We need to return an error - this one belongs
		// to the mem manager... but don't know what else to use.
		return memErrNotEnoughSpace;
	}
	debug_CapIns = debug_CapBuf;
	debug_CapEnd = debug_CapBuf + CAPTURE_SIZE - 1;
#endif

	return error;
}

/*
 * Stop Palm Application...
 *
 */

static void
stopApp(void)	//SEG2
{
	UI_mainOnScreen = 0;
	UI_mainDrawingOK = 0;
	Screen_Stop();

	Serial_ClosePort();
	Serial_Fini();
	MO_Fini();
	Screen_Fini();

	PrefSetAppPreferences((UInt32) 'POrb', 1, CurrentPrefVersion,
						  &AppPrefs, sizeof(AppPrefs), true);

#if DEBUG_CAPTURE
	if (debug_CapBuf)
	{
		MemPtrFree(debug_CapBuf);
		debug_CapBuf = NULL;
	}
#endif

	FrmSaveAllForms();
	FrmCloseAllForms();
}



static void
processKey(UInt16 ch)
{
	switch (ch)
	{
		case lowBatteryChr:
			// Tell user of change in status...
			powerManagement(true);
			break;	// pass key on..

#if DEBUG_CAPTURE
		case 'l':
		case 'L':
			// Time to feed the captured data back down the serial port!
			{
				// Mark beginning
				if (ch=='l')debug_sendBuf((UInt8 *) "|=START=|", 9);
				if (debug_CapWrapped)
					debug_sendBuf(debug_CapIns,
								  debug_CapEnd - debug_CapIns + 1);
				if (debug_CapIns - debug_CapBuf > 0)
					debug_sendBuf(debug_CapBuf, debug_CapIns - debug_CapBuf);
				// Mark end
				if (ch=='l')debug_sendBuf((UInt8 *) "|=END=|", 7);
			}
			break;
#endif

	}

	if (AppPrefs.sendGraffiti &&
	    ((ch>='a' && ch <='z') || (ch>='A' && ch<='Z')))
	{
				MO_SendKey(ch);
	}
}

#define ExitingMainWindow ((FrmGetFormPtr(MainFormID)==NULL) || \
						(event->data.winExit.exitWindow == \
							FrmGetWindowHandle(FrmGetFormPtr(MainFormID))))

#define EnteringMainWindow ((FrmGetFormPtr(MainFormID)!=NULL) && \
						(event->data.winEnter.enterWindow == \
							FrmGetWindowHandle(FrmGetFormPtr(MainFormID))) && \
						(event->data.winEnter.enterWindow == \
							FrmGetWindowHandle(FrmGetFirstForm())))


static void
specialPreProcessing(EventPtr event)
{
	switch (event->eType)
	{
		case keyDownEvent:
			processKey(event->data.keyDown.chr);
			break;


		// This event occurs when a FrmSetActiveForm is called.
		// It is sent to the old form.
		case winExitEvent:
			if (ExitingMainWindow)
			{
				UI_mainOnScreen = 0;
				Screen_Stop();
			}
			break;
			
		case winEnterEvent:
			if (!EnteringMainWindow)
			{
				// must be entering something else then...
				UI_mainOnScreen = 0;
				Screen_Stop();
			}

		default:
			break;
	}

}

static void
specialPostProcessing(EventPtr event)
{
	switch (event->eType)
	{
		// Pre PalmOS 3.5: This event occurs when a FrmSetActiveForm is called.
		// PalmOS 3.5+: This event occurs after the form is drawn.
		case winEnterEvent:

			// We use this because there are cases when
			// we don't get frmUpdateEvent/frmOpenEvent, like
			// when menus close.
			//case frmUpdateEvent:
			if (EnteringMainWindow)
			{
				if (UI_mainDrawingOK)
				{
					UI_mainOnScreen = 1;

					FormType *frm = FrmGetFormPtr(MainFormID);
					ASSERT(frm, "MainFormID not loaded");
					if (frm)
					{
						UI_showPowerStatus(frm, lowBattery);

						Screen_SetColorMode(AppPrefs.colorMode);
						Screen_Start();
					}
				}
			}
			break;

		default:
			break;
	}

	readSerial();
}


static void
readSerial(void)
{
	UInt32 size = 0;
	UInt8 *buffer;
	size = Serial_ReadPort(&buffer);

	if (size > 0)
	{
		ASSERT(buffer, "buffer is NULL");

#if DEBUG_CAPTURE
		if (debug_CapIns + size > debug_CapEnd)
		{	// it will need to wrap around
			UInt32 left = debug_CapEnd - debug_CapIns + 1;
			if (left)
				MemMove(debug_CapIns, buffer, left);
			if (size - left)
				MemMove(debug_CapBuf, buffer + left, size - left);
			debug_CapIns = debug_CapBuf + size - left;
			debug_CapWrapped = 1;
		}
		else
		{
			MemMove(debug_CapIns, buffer, size);
			debug_CapIns += size;
		}
#endif

		// Some data has arrived so reset the idle timer so we don't
		// switch off.
		powerManagement(false);

		// Send the data to the MatrixOrbital emulation code...
		MO_ProcessInput(buffer, size);

		// We used to use less than the full buffer and hope that the remains
		// would be sent next time with new data appended - but it seems
		// it doesn't always work. See Note 2 in NOTES.
		Serial_CloseData(buffer, size);
	}
}

static void
eventLoop(void)
{
	EventType event;
	UInt16 error;
	UInt32 timeOut = SysTicksPerSecond();

	timeOut /= 50;

	do
	{

		/* Get the next available event. */
		EvtGetEvent(&event, timeOut);

		specialPreProcessing(&event);

		/* Give the system a change to handle the event. */
		if (!SysHandleEvent(&event))
		{
			/* Give the menu bar a chance to update and
			 * handle the event. 
			 */
			if (!MenuHandleEvent(NULL, &event, &error))
			{
				/* Give the application a chance to handle
				 * the event. 
				 */
				if (!UI_ApplicationHandleEvent(&event))
				{
					/* Let the form object provide default
					 * handling of the event. 
					 */
					FrmDispatchEvent(&event);
				}
			}
		}

		specialPostProcessing(&event);

#if DEBUG
		// This being called rather a lot
		// Seems pointless - but lets see...
		DebugCheckStack();
#endif

	}
	while (event.eType != appStopEvent);
}

/*
 * Check ROM Version
 *
 */

/* romVersionCompatible based on code from Address source code from Palm */
static Err
romVersionCompatible(UInt16 launchFlags)
{
	UInt32 romVersion = 0;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < sysMakeROMVersion(2, 0, 0, sysROMStageRelease, 0))
	{
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
			== (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
		{

			FrmAlert(RomIncompatibleAlert);

			/* Deal with v1.0 */
			if (romVersion < 0x02000000)
				AppLaunchWithCommand(sysFileCDefaultApp,
									 sysAppLaunchCmdNormalLaunch, NULL);
		}

		return sysErrRomIncompatible;
	}

	return 0;
}


/*
 *  Pilot Main Start
 *
 */
UInt32
PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error = 0;

	/* Check the ROM version */
	error = romVersionCompatible(launchFlags);
	if (error)
		return error;


	switch (cmd)
	{
		case sysAppLaunchCmdNormalLaunch:
			error = startApp();
			if (error == errNone)
			{
				// App has initialised fine
				Serial_OpenPort();
				FrmGotoForm(MainFormID);
				eventLoop();
			}
			stopApp();
			break;
	}

	return error;
}
