/******************************************************************************
 *
 *  PalmOrb - Emulate a Matrix Orbital LCD Display with Keypad
 *  Copyright (C) 2004  Chris Lansley  <palmorb@lansley.me.uk>
 *  Copyright (C) 2001  David Douthitt 
 *  Copyright (C) 2000  J Robert Ray
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  $Source: /cvsroot/palmorb/palmorb/screen.cpp,v $
 *  $Revision: 1.16 $ $Date: 2004/11/02 21:13:32 $
 *
 ******************************************************************************/

    /******************************************************************/
    /* This file contains mostly the code for handing the LCD screen. */
    /******************************************************************/

#include <PalmOS.h>

#include "palmorb.h"
#include "screen.h"
#include "MatrixOrbital.h"
#include "debug.h"
#include "types.h"



/* on screen borders defines 
 * for both lcd and special chars
 */
#define BORDER_SPACING 2
#define BORDER_PADDING (BORDER_SPACING+1)

/* font defines */
#define FW_CHARS_PER_ROW 16 //chars per row in fontwin
#define FONT_START_CHAR 32

/* defines for special chars (user defined chars) */
#define SCHAR_CELL_SPACING 2
#define SCHAR_FULL_FONTCELL_WIDTH (FONTCELL_WIDTH + (SCHAR_CELL_SPACING * 2)+1)
#if NEWSCREEN
#define SCHAR_OFFSET_X 42
#define SCHAR_OFFSET_Y 102 //147
#else
#define SCHAR_OFFSET_X 32
#define SCHAR_OFFSET_Y 66
#endif
#define SCHAR_CODE_OFFSET 128
   /*for within fontwin*/
#define SCHAR_FW_OFFSET_X    (FONTCELL_WIDTH * 6)
#define SCHAR_FW_OFFSET_Y(y) ((y) * FONTCELL_HEIGHT)

/* defines for the LCD */
#define LCD_X(x) ((x) * (FONTCELL_WIDTH + lcdGap))
#define LCD_Y(y) ((y) * (FONTCELL_HEIGHT+ lcdGap))
#define LCDTopLeftX(w)	((160-(w))/2)
#if NEWSCREEN
#define LCDTopLeftY(h)  (18/*height of header*/+ (100/*start of keys*/-18/*height of header*/-(h))/2)
#else
#define LCDTopLeftY(h)  (21)
#endif


/* Color defines */
#define COLOR_FOREGROUND 1
#define COLOR_BACKGROUND 0


/*Exported Variables*/
WinHandle Screen_Fontwin = NULL; // for palmorb.cpp to show font map
Boolean Screen_ColorPossible = false; // so the menu can be disabled

/*Exported Functions*/
Err Screen_Init(void);
void Screen_Fini(void);
void Screen_UpdateLcd(void);
void Screen_Start(void);
void Screen_Stop(void);
void Screen_setOutput(UInt8 num, Boolean set);
void Screen_PutChar (UInt8 chr, UInt16 x, UInt16 y);
void Screen_DefineChar (UInt8 ch, const UInt8 *str);
void Screen_PutNCharsAt (Char *str, UInt16 n, UInt16 x, UInt16 y);
void Screen_Clear(void);
void Screen_DoAutoscroll (void);
void Screen_SetColorMode(UInt32 type);
void Screen_PrefsChanged(void);
void Screen_SetConfig(UInt16 width, UInt16 height, 
		UInt16 numoutputs, UInt16 numkeys);

/*File Globals*/
static WinHandle bufwin = NULL;
static MUInt16 lcdHeight=0;
static MUInt16 lcdWidth=0;
static MUInt16 winHeight=0;
static MUInt16 winWidth=0;
static MUInt8 numOutputs=0;
static MUInt8 numKeys=0;
static MUInt16 lcdGap=1;
static Boolean enabled=0;
static MUInt8 colorMode=0xFFFF; // undefined.
static UInt8 colorCharForeIndex;
static UInt8 colorCharBackIndex;
static UInt8 colorLcdBackIndex;
static Boolean customNew[NUM_SPECIAL_CHARS]; // has a custom char changed?
static UInt8 customCharData[NUM_SPECIAL_CHARS][FONTCELL_HEIGHT];


	// textScreen contains a transformed version of the LCD
static UInt8  *textScreen; 
static UInt8  *old_textScreen; 
static UInt32 textScreenSize;

static Boolean modified = 0;
static Boolean drawingOK = false;
static UInt8 gpo[8] =  { 0, 0, 0, 0, 0, 0, 0, 0 }; // General Propose Outputs
static Boolean noWinDrawPixel = true;
static Boolean noWinSetColor = true;
static Boolean noWinPalette = true; 
static Boolean noWinDrawState = true;  // noWin[Push|Pop]DrawState
static Boolean noWinScreenMode = true; 
static Boolean slashCorrected = 0;



/*Local Functions*/
static RectangleType *setFWCellRectangle (RectangleType *srect, UInt8 ch);
static void defineChar(UInt16 start_x, UInt16 start_y, const UInt8 *str);
static inline UInt8 charTransform(UInt8 chr);
static void myPushDrawState(void);
static void myPopDrawState(void);
static void setOutputs (void);
static void hideAllKeys (void);
static void showOutputs(void);
static void screen_ShowKeys(void);
static void screen_draw(void);
static void drawSpecialCharBorders(void);
static void drawLcdBorder(void);
static IndexedColorType mySetColor(Boolean foreground, UInt8 index);
static void fontWin_init(void);
static void correctSlash(void);
static void resize(MUInt8 padCharCells, MUInt16 newLcdWidth = lcdWidth,
	   	           MUInt16 newLcdHeight = lcdHeight);

Err
Screen_Init(void)
{
	Err error = errNone;
	UInt32 romVersion=0;

	ASSERT(!enabled, "we're already enabled");


	// Check if WinDrawPixel is available - is on OS 3.5 +
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion); 

	if (romVersion >= sysMakeROMVersion(3,5,0,sysROMStageRelease,0) )
	{
		noWinDrawPixel = false;
		noWinSetColor = false;
		noWinPalette = false; 
		noWinDrawState = false; 
		noWinScreenMode = false; 
	}

	// Decide if we want to enable color support
	if (!noWinScreenMode)
	{
		UInt32 depth;
		Err err;
		Boolean color;

		err=WinScreenMode (winScreenModeGet, NULL, NULL, &depth, &color);
		ASSERT(err==errNone, "WinScreenMode failed");
		if ((err == errNone) && (depth>=8) && (color))
			Screen_ColorPossible = true;
	}

    // LCD Font "Storage" Window
    Screen_Fontwin = WinCreateOffscreenWindow(FW_WIDTH, FW_HEIGHT,
                                        nativeFormat, &error);
	if (error!=errNone)
	{
		ASSERT(error==errNone, "WinCreateOffscreenWindow errored");
		FrmAlert(FailedToCreateFontWinAlert);
	}
	else
	{
		fontWin_init();
	}
    
	if (error==errNone)
		enabled=1;

	return error;
}

void
Screen_SetConfig(UInt16 width, UInt16 height,
		UInt16 numoutputs, UInt16 numkeys)
{
	ASSERT1(enabled, "Init hasn't been called", enabled);

	ASSERT(width, "width is 0");
	ASSERT(height, "height is 0");

	numOutputs=numoutputs;
	numKeys=numkeys;

	// resize data LCD structures if required.
	resize(AppPrefs.padCharCells, width, height); 

	if (enabled == 1) // if we have been inited (and not turned off)
		enabled=2;    // then it's ok to turn drawing on

	if (drawingOK)
		screen_draw();
}

static void
resize(MUInt8 newLcdGap, MUInt16 newLcdWidth = lcdWidth,
	   	MUInt16 newLcdHeight = lcdHeight)
{
	MUInt8 oldLcdGap = lcdGap;
	MUInt16 oldWinWidth = winWidth;
	MUInt16 oldWinHeight = winHeight;
	MUInt16 oldLcdWidth = lcdWidth;
	MUInt16 oldLcdHeight = lcdHeight;


	// If the size has changed then recreate it
	if ((!bufwin) ||
		   	(newLcdGap != oldLcdGap) ||
			(newLcdWidth != oldLcdWidth) ||
			(newLcdHeight != oldLcdHeight))
	{
		// Set up new globals with new values
		lcdWidth=newLcdWidth;
		lcdHeight=newLcdHeight;
		lcdGap = newLcdGap;
		winWidth=((FONTCELL_WIDTH + newLcdGap) * newLcdWidth) - newLcdGap;
		winHeight=((FONTCELL_HEIGHT + newLcdGap) * newLcdHeight) - newLcdGap;

		if (bufwin)
		{
			/*
			 * Erase the old window from the screen
			 */
			FormType *main = FrmGetFormPtr(MainFormID);
			ASSERT(main, "MainForm not loaded");

			if (main)
			{
				WinHandle oldwin = WinSetDrawWindow (FrmGetWindowHandle(main));
				//myPushDrawState();
				RectangleType rect;
				rect.topLeft.x = LCDTopLeftX(oldWinWidth) - BORDER_PADDING;
				rect.topLeft.y = LCDTopLeftY(oldWinHeight) - BORDER_PADDING;
				rect.extent.x = oldWinWidth + (BORDER_PADDING * 2);
				rect.extent.y = oldWinHeight + (BORDER_PADDING * 2);
				WinEraseRectangle(&rect, 0);
				WinEraseRectangleFrame(simpleFrame, &rect);
				//myPopDrawState();
				WinSetDrawWindow (oldwin);
			}

			/*
			 * Free old window
			 */
			WinDeleteWindow (bufwin, false);
			bufwin=NULL;
		}


		Err error;
		// bufwin: Offscreen LCD Display Buffer
		bufwin = WinCreateOffscreenWindow(winWidth, winHeight,
											nativeFormat, &error);
		ASSERT((error==errNone) && (bufwin), "WinCreateOffscreenWindow errored");
		if (bufwin)
		{
			// Set it's background color.
			WinHandle oldwin = WinSetDrawWindow (bufwin);
			//myPushDrawState();
			IndexedColorType old;
			old = mySetColor(COLOR_BACKGROUND, colorLcdBackIndex);
			WinEraseWindow ();
			mySetColor(COLOR_BACKGROUND, old);
			//myPopDrawState();
			WinSetDrawWindow (oldwin);
		}
	   	else if (error!=errNone)
		{
			// This is bad news...
			
			FrmAlert(FailedToCreateBufWinAlert);

			if (bufwin)
				WinDeleteWindow (bufwin, false);
			bufwin=NULL;

			// disable the Screen code.
			drawingOK=0;
			enabled=0;
		}

		// If the size of the text screen has changed then realloc...
		if ((!textScreen) || 
				(!old_textScreen) ||
				(oldLcdWidth != newLcdWidth) ||
				(oldLcdHeight != newLcdHeight))
		{
			if (textScreen)
			{
				MemPtrFree(textScreen);
				textScreen=NULL;
			}

			if (old_textScreen)
			{
				MemPtrFree(old_textScreen);
				old_textScreen=NULL;
			}

			textScreenSize = newLcdWidth*newLcdHeight;
			textScreen = (UInt8 *)MemPtrNew(textScreenSize);
			old_textScreen = (UInt8 *)MemPtrNew(textScreenSize);

			if (textScreen)
			{
								   //12345678901234567890
				UInt8 initText1[] = "PalmOrb";
				UInt8 initText2[] = "version " STRINGIFY(VERSION);
#if DEBUG
				UInt8 initText3[] = "DEBUG BUILD";
#elif ALPHARELEASE
				UInt8 initText3[] = "Alpha Release";
#elif BETARELEASE
				UInt8 initText3[] = "Beta Release";
#elif EXPERIMENTALRELEASE
				UInt8 initText3[] = "Experimental Build";
#else
				UInt8 initText3[] = "";
#endif

				// Initalize the initial screen.
				MemSet(textScreen, textScreenSize, ' ');

				if (lcdHeight >= 2) // need at least two lines
				{
					UInt8 *firstLine=textScreen + (lcdHeight/2-1)*newLcdWidth;
					UInt8 *secondLine=firstLine+newLcdWidth;

					if (sizeof(initText1)-1 <= newLcdWidth)
						MemMove(firstLine+(newLcdWidth-
												(sizeof(initText1)-1))/2,
								initText1, sizeof(initText1)-1);

					if (sizeof(initText2)-1 <= newLcdWidth)
						MemMove(secondLine+(newLcdWidth-
												(sizeof(initText2)-1))/2,
								initText2, sizeof(initText2)-1);

					if (newLcdHeight>2)
					{ // there's room for a third line
						UInt8 *thirdLine = secondLine+newLcdWidth;

						if ((sizeof(initText3) > 1) &&
								(sizeof(initText3)-1 <= newLcdWidth))
							MemMove(thirdLine+(newLcdWidth-
												(sizeof(initText3)-1))/2,
									initText3, sizeof(initText3)-1);
					}
				}
			}
			else
			{
				// This is bad news...
				FrmAlert(FailedToCreateLcdAlert);
				ASSERT(textScreen, "Failed to allocate memory for lcd");
				textScreenSize=0;
				// Disable the screen code.
				drawingOK=0;
				enabled=0;
			}
		}

		/* Do this even if the text screen size hasn't changed...
		 * because the bufwin is empty and need redrawing.
		 */
		if (old_textScreen)
		{
			ASSERT(textScreen, "old_textScreen is allocated but textScreen isn't");
			// Fill it up with something that will never match whats
			// on textScreen so it's forced to update
			MemSet(old_textScreen, textScreenSize, 254);
		}
		else
		{
			ASSERT(0, "Failed to allocate for old_textScreen");
			
			// we could live without it ... but why bother
			FrmAlert(FailedToCreateOLcdAlert);
			ASSERT(old_textScreen, "Failed to allocate memory for lcd");
			textScreenSize=0;
			// Disable the screen code.
			drawingOK=0;
			enabled=0;
		}
	}
	else
	{
		ASSERT(bufwin, "bufwin is NULL!");
		ASSERT(textScreen, "textScreen is NULL");
		ASSERT(old_textScreen, "old_textScreen is NULL");
	}
}

void
Screen_Fini(void)
{

    if (bufwin)
	{
        WinDeleteWindow (bufwin, false);
		bufwin=NULL;
	}

    if (Screen_Fontwin)
	{
        WinDeleteWindow (Screen_Fontwin, false);
		Screen_Fontwin=NULL;
	}

	if (textScreen)
	{
		MemPtrFree(textScreen);
		textScreen=NULL;
	}

	if (old_textScreen)
	{
		MemPtrFree(old_textScreen);
		old_textScreen=NULL;
	}

	drawingOK=0;
	enabled=0;
}


//CJL: could also -32 here also?
static inline UInt8 
charTransform(UInt8 chr)
{
	// [some chars will have been handled already and not get here:
	//   254 - command character
	//   10  - line feed
	//   13  - carriage return
	//   8   - backspace
	//   12  - clear screen
	//  ]
	ASSERT((chr!=254) && (chr!=10) && (chr!=13) && (chr!=8) && (chr!=12),
			"control character unexpected");

	// deal with most common cases first.
	if (((chr>=32) && (chr<=127)) || (chr==255))
		return chr;
	
	// Currently user defined chars (0-7)
	// are mapped to SCHAR_CODE_OFFSET in the font.
	if (chr<8)
		return chr + SCHAR_CODE_OFFSET;
	
	// For ease of later processing map all empty
	// chars to the space char.
	if (((chr>=9) && (chr<=32)) ||
		((chr>=128) && (chr<=160)))
		return 32;


	// The remaining chars are less commonly used
	// but remain untransformed.
	ASSERT((chr>=161) && (chr<=253), "non handled case");
	return chr;
}

static void
DrawSpecialChar(WinHandle ToWindow, UInt8 specChr) 
{
	RectangleType srect;
	UInt8 CharX, CharY;
	UInt8 chr;

	ASSERT(specChr<NUM_SPECIAL_CHARS, "specialchar out of range");
	
	// Map the special char into our font
	chr = specChr + SCHAR_CODE_OFFSET;

	CharX = SCHAR_OFFSET_X + (SCHAR_FULL_FONTCELL_WIDTH * specChr);
	CharY = SCHAR_OFFSET_Y;

	ASSERT(Screen_Fontwin, "font is NULL!");
	setFWCellRectangle (&srect, chr);
	WinCopyRectangle (Screen_Fontwin, ToWindow, &srect, CharX, CharY, winPaint);
}



void 
Screen_Stop(void)
{
	drawingOK=false;
}

static void
myPushDrawState(void)
{
	if (!noWinDrawState)
		WinPushDrawState();
}

static void
myPopDrawState(void)
{
	if (!noWinDrawState)
		WinPopDrawState();
}


/* NB: Do not call until main form is drawn and active.
 * This function uses WinPushState/WinPopState which seem to need
 * an active Draw window to function. 
 */
void
Screen_SetColorMode(UInt32 type)
{
#define LCD_COLORS 3
	Err err;
	RGBColorType colors[LCD_COLORS];
	FormType *mainForm;
	Boolean updatePalette = false;

	// check if the mode has changed
	if (type == colorMode)
		return;
	colorMode=type;

	mainForm = FrmGetFormPtr(MainFormID);
	ASSERT(mainForm, "Main not loaded");
	if (!mainForm)
		return;

	ASSERT(type<=COLORMODE_MAX, "color mode too high");

	if ((!Screen_ColorPossible) || (type>COLORMODE_MAX))
		return;

	ASSERT(!noWinPalette, "noWinPalette but ColorPossible!");

	switch (type)
	{
		case COLORMODE_GREEN: // Green LCD
			colorCharBackIndex = 252;
			colorCharForeIndex = 253;
			colorLcdBackIndex = 254;

			updatePalette = 1;

			// Char foreground - black
			colors[1].r = 0;
			colors[1].g = 0;
			colors[1].b = 0;

			// Char background
			colors[0].r = 112;
			colors[0].g = 192;
			colors[0].b = 0;

			// LCD background
			colors[2].r = 128;
			colors[2].g = 208;
			colors[2].b = 0;

			break;

		case COLORMODE_BLUE: // Blue LCD
			colorCharBackIndex = 252;
			colorCharForeIndex = 253;
			colorLcdBackIndex = 254;

			updatePalette = 1;
			
			// Char foreground - white
			colors[1].r = 255;
			colors[1].g = 255;
			colors[1].b = 255;

			// Char background
			colors[0].r = 2;
			colors[0].g = 51;
			colors[0].b = 252;

			// LCD background
			colors[2].r = 3;
			colors[2].g = 83;
			colors[2].b = 255;
			break;

		case COLORMODE_CYAN: // CYAN LCD
			colorCharBackIndex = 252;
			colorCharForeIndex = 253;
			colorLcdBackIndex = 254;

			updatePalette = 1;
			
			// Char foreground - cyan
			colors[1].r = 0;
			colors[1].g = 255;
			colors[1].b = 225;

			// Char background
			colors[0].r = 1;
			colors[0].g = 63;
			colors[0].b = 53;

			// LCD background
			colors[2].r = 0;
			colors[2].g = 0;
			colors[2].b = 0;
			break;

		case COLORMODE_RED: // RED LCD
			colorCharBackIndex = 252;
			colorCharForeIndex = 253;
			colorLcdBackIndex = 254;

			updatePalette = 1;
			
			// Char foreground
			colors[1].r = 252;
			colors[1].g = 0;
			colors[1].b = 1;

			// Char background
			colors[0].r = 62;
			colors[0].g = 0;
			colors[0].b = 5;

			// LCD background
			colors[2].r = 0;
			colors[2].g = 0;
			colors[2].b = 0;
			break;


		case COLORMODE_NONE: // Disable color
		default:
			colorCharBackIndex = UIColorGetTableEntryIndex(UIFormFill);
			colorCharForeIndex = UIColorGetTableEntryIndex(UIObjectForeground);
			colorLcdBackIndex = UIColorGetTableEntryIndex(UIFormFill);

			updatePalette = 0;
	}
		   
	if (updatePalette)
	{
		WinHandle oldwin = WinSetDrawWindow(FrmGetWindowHandle(mainForm));
		// The manual doesn't say that WinPalette changes the fg/bg colors
		// but it does seem to...
		myPushDrawState();
		err = WinPalette ( winPaletteSet, 255-LCD_COLORS, LCD_COLORS, colors);
		myPopDrawState();
		ASSERT(err==errNone, "WnPalette failed");
		WinSetDrawWindow(oldwin);
	}

	fontWin_init(); // This will update the colors in the font window
	
	ASSERT(bufwin, "bufwin is NULL");

	if (bufwin)
	{
		// Updates the bufwin's background color
		WinHandle oldwin = WinSetDrawWindow (bufwin);
		//myPushDrawState();
		IndexedColorType old;
		old = mySetColor(COLOR_BACKGROUND, colorLcdBackIndex);
		WinEraseWindow ();
		mySetColor(COLOR_BACKGROUND, old);
		//myPopDrawState();
		WinSetDrawWindow (oldwin);
	}
	
	// Fill old_textScreen up with something that will never match whats
	// on textScreen so it's forced to update
	if (old_textScreen)
		MemSet(old_textScreen, textScreenSize, 254);

	if (drawingOK)
		screen_draw();
}

/* Set the LCD colors up:
 *   foreground == COLOR_FOREGROUND ==> set foreground, other background
 */
static IndexedColorType 
mySetColor(Boolean foreground, UInt8 index)
{
	IndexedColorType old = 0;
	
	if (!noWinSetColor && Screen_ColorPossible)
	{ // We have WinSet[Fore|Back]Color...

		if (foreground == COLOR_FOREGROUND)
			old = WinSetForeColor(index);
		else
			old = WinSetBackColor(index);
	}

	return old; // when color isn't possible then this value
   				// doesn't matter anyway.
}

static void
correctSlash(void)
{
	if (AppPrefs.correctSlash)
	{
		UInt8 charNum = '\\' - FONT_START_CHAR;
			// ----- 0x00
			// #---- 0x10
			// -#--- 0x08
			// --#-- 0x04
			// ---#- 0x02
			// ----# 0x01 
			// ----- 0x00
			// ----- 0x00
		defineChar(FONTCELL_WIDTH*(charNum/FW_CHARS_PER_ROW),
				   	FONTCELL_HEIGHT*(charNum%FW_CHARS_PER_ROW),
				   	(UInt8 *)"\x00\x10\x08\x04\x02\x01\x00\x00");
		slashCorrected = 1;
	}
}

void
Screen_PrefsChanged(void)
{
	if (AppPrefs.correctSlash == slashCorrected)
		return;

	if (AppPrefs.correctSlash && !slashCorrected)
		correctSlash();
	else if (!AppPrefs.correctSlash && slashCorrected)
		fontWin_init();
	
	// Fill old_textScreen up with something that will never match whats
	// on textScreen so it's forced to update
	if (old_textScreen)
		MemSet(old_textScreen, textScreenSize, 254);

	if (drawingOK)
		screen_draw();
}

static void
fontWin_init(void)
{
	MemHandle font = DmGetResource(bitmapRsc, FontBitmapID);

	if (font)
	{
		BitmapPtr fontp = (BitmapPtr) MemHandleLock(font);

		if (fontp)
		{
			WinHandle oldwin;

			ASSERT(Screen_Fontwin, "font is NULL!");
			oldwin 	= WinSetDrawWindow(Screen_Fontwin);

			//myPushDrawState();

			IndexedColorType oldf, oldb;
			oldb=mySetColor(COLOR_BACKGROUND, colorCharBackIndex);
			oldf=mySetColor(COLOR_FOREGROUND, colorCharForeIndex);

			WinDrawBitmap(fontp, 0, 0);

			mySetColor(COLOR_BACKGROUND, oldb);
			mySetColor(COLOR_FOREGROUND, oldf);

			//myPopDrawState();
			WinSetDrawWindow(oldwin);

			MemHandleUnlock(font);

			//We have reset the slash back to a yen.
			slashCorrected = 0;

			correctSlash();

			// This process has wiped our custom chars which
			// are drawn onto the fontwin...  Restore them.
			for (UInt16 i = 0; i < NUM_SPECIAL_CHARS; i++)
				Screen_DefineChar (i, customCharData[i]);
		}
		else
		{
			ASSERT(0, "MemHandleLock failed");
			return;
		}

		DmReleaseResource(font);
	}
	else
	{
		ASSERT(0, "DmGetResource failed");
		return;
	}
}

void
Screen_Start(void)
{
	ASSERT(FrmGetFormPtr(MainFormID), "MainForm not loaded");

	// ensure that init and setconfig have been called
	if (enabled!=2)
		return;

	ASSERT(bufwin, "bufwin is NULL!");
	ASSERT(Screen_Fontwin, "Screen_Fontwin is NULL!");
	ASSERT(textScreen, "textScreen is NULL");
	ASSERT(old_textScreen, "old_textScreen is NULL");

	// resize data LCD structures if required.
	resize(AppPrefs.padCharCells);

	drawingOK=true;
	screen_draw();
}

/* 
 * Setup srect to point to char ch within FontWin
 */
static RectangleType *
setFWCellRectangle (RectangleType *srect, UInt8 ch) 
{
    UInt8 fontchar;

    fontchar = ch;

    // First two columns of the font chart are "missing";
    // Special characters are placed to the right of the
    // column containing p-q-r-s; others are blank.

	ASSERT(ch>=FONT_START_CHAR, "ch not transformed");

	fontchar -= FONT_START_CHAR;

    // define topLeft(x,y) of font character
    srect->topLeft.x = (fontchar / FW_CHARS_PER_ROW) * FONTCELL_WIDTH;
    srect->topLeft.y = (fontchar % FW_CHARS_PER_ROW) * FONTCELL_HEIGHT;

    // size of font character
    srect->extent.x = FONTCELL_WIDTH;
    srect->extent.y = FONTCELL_HEIGHT;

    return srect;
}

/* 
 * This draws the main screen.
 * Includes: keyboard, outputs, LCD, special characters display.
 */
static void
screen_draw(void)
{
	ASSERT(drawingOK, "Drawing is not ok!");

	setOutputs();
	showOutputs(); // Trims those to be displayed and those not...
	screen_ShowKeys();
	modified=true;
	drawSpecialCharBorders(); // also draws SC backgrounds & clears it.
	drawLcdBorder(); // also draws LCD background
	Screen_UpdateLcd();
}

static void 
drawLcdBorder(void)
{
	RectangleType border;
	FormType *mainForm;

    mainForm=FrmGetFormPtr(MainFormID);
	ASSERT(mainForm, "main not loaded");
	if (!mainForm)
		return;


   	border.topLeft.x = LCDTopLeftX(winWidth) - BORDER_PADDING;
	border.topLeft.y = LCDTopLeftY(winHeight) - BORDER_PADDING;
	border.extent.x = winWidth  + (BORDER_PADDING * 2);
	border.extent.y = winHeight + (BORDER_PADDING * 2);

	ASSERT(drawingOK, "Drawing is not ok!");

	WinHandle oldwin=WinSetDrawWindow(FrmGetWindowHandle(mainForm));

	// If color is possible then make sure the space between the LCD
	// and border is also in the right color.
	if (Screen_ColorPossible)
	{
		//myPushDrawState();
		IndexedColorType old;
		old = mySetColor(COLOR_FOREGROUND, colorLcdBackIndex);
		WinDrawRectangle(&border, 0);
		mySetColor(COLOR_FOREGROUND, old);
		//myPopDrawState();
	}
	
	// Draw frame around LCD display
	WinDrawRectangleFrame(simpleFrame, &border);

	WinSetDrawWindow(oldwin);
}

/* If showing special chars is:
 *   enabled: draw background and borders
 *   (recently) disabled: clear area.
 */
static void 
drawSpecialCharBorders(void)
{
	static Boolean showing = false;
	RectangleType border; 
	FormType *mainForm;

	if (!AppPrefs.showCustomChars && !showing)
		return;

    mainForm=FrmGetFormPtr(MainFormID);
	ASSERT(mainForm, "main not loaded");
	if (!mainForm)
		return;


	// Setup size of border
	border.extent.x = FONTCELL_WIDTH + (BORDER_SPACING * 2);
	border.extent.y = FONTCELL_HEIGHT + (BORDER_SPACING * 2);

	// Setup Y pos of border (X pos changes)
	border.topLeft.y = SCHAR_OFFSET_Y - BORDER_SPACING;

	ASSERT(drawingOK, "Drawing is not ok!");

	WinHandle oldwin=WinSetDrawWindow(FrmGetWindowHandle(mainForm));
	IndexedColorType old=0;
	if (AppPrefs.showCustomChars)
		old = mySetColor(COLOR_BACKGROUND, colorLcdBackIndex);
	//else use background color to erase..

	for (UInt32 i=0; i<NUM_SPECIAL_CHARS; i++)
	{
		border.topLeft.x = SCHAR_OFFSET_X 
							+ (SCHAR_FULL_FONTCELL_WIDTH * i)
							- BORDER_SPACING;

		// If color is enabled then make sure the space between the Custom char
		// and border is also in the right color. (also used to erase when
		// disabled).
		WinEraseRectangle(&border, 0);

		if (AppPrefs.showCustomChars)
			WinDrawRectangleFrame(simpleFrame, &border);
		else
			WinEraseRectangleFrame(simpleFrame, &border);

	}

	if (AppPrefs.showCustomChars)
		mySetColor(COLOR_BACKGROUND, old);
	WinSetDrawWindow(oldwin);

	showing = AppPrefs.showCustomChars;
}


/*
 * This updates only the LCD and special character display.
 */
void
Screen_UpdateLcd(void)
{
	UInt16 ch, i;

	// No modifications: quick exit
	if (!modified) 
	{
		return;
	}

	// Don't update the screen if main form isn't displayed
	// (or menu is shown)
	if (! drawingOK) 
	{
		return;
	}

	FormType *frm = FrmGetFormPtr(MainFormID);
	ASSERT(frm, "frm is NULL");
	if (!frm)
		return;


	ASSERT(bufwin, "bufwin is NULL!");
	ASSERT(Screen_Fontwin, "Screen_Fontwin is NULL!");

	ASSERT(textScreen, "textScreen is NULL");
	ASSERT(old_textScreen, "old_textScreen is NULL");

	UInt8 *p = textScreen;
	UInt8 *q = old_textScreen;
	for (UInt16 y = 0; y < lcdHeight; y++) 
	{
		for (UInt16 x = 0; x < lcdWidth; x++) 
		{
			ch = *p;
			ASSERT(ch>=FONT_START_CHAR, "char not transformed");

			if ( (*q != ch) ||  // The character has changed
				((ch>=SCHAR_CODE_OFFSET) && // OR it's a custom char
				 (ch<=SCHAR_CODE_OFFSET+7) && // whose def has changed.  
					customNew[ch-SCHAR_CODE_OFFSET]))     
			{
				RectangleType srect;

				setFWCellRectangle(&srect, ch);
				WinCopyRectangle (Screen_Fontwin, bufwin, &srect,
					            	LCD_X(x), LCD_Y(y), winPaint);
				*q = *p;
			}

			p++;
			q++;
		}
	}
	
	// This is an oversized window; the LCD display is in the top left of it.
	// The window extends down and to the right edge...
	RectangleType rect;
	rect.topLeft.x=0;
	rect.topLeft.y=0;
	rect.extent.x=winWidth;
	rect.extent.y=winHeight;

	// Draw LCD window
    WinCopyRectangle (bufwin, FrmGetWindowHandle(frm) , &rect, 
						LCDTopLeftX(winWidth), LCDTopLeftY(winHeight), winPaint);


	for (i = 0; i < NUM_SPECIAL_CHARS; i++)
		customNew[i]=0;

	if (AppPrefs.showCustomChars)
	{
		for (i = 0; i < NUM_SPECIAL_CHARS; i++)
			DrawSpecialChar(FrmGetWindowHandle(frm), i);
	}

	modified = 0;
}





void
Screen_DoAutoscroll (void) 
{
	// ensure that init and setconfig have been called
	ASSERT(enabled==2, "Init and SetConfig not yet called?");
	if (enabled!=2)
		return;

	ASSERT(textScreen, "textScreen is NULL");

	// move lines up one line (scroll up)
	MemMove(textScreen, textScreen+lcdWidth, lcdWidth*(lcdHeight-1));

	// initialize bottom line to blanks
	MemSet(textScreen + lcdWidth*(lcdHeight-1), lcdWidth, ' ');

	ASSERT(charTransform(' ')==' ', "space is being transformed");

	modified = 1;
}



void
Screen_Clear(void) 
{
	// ensure that init and setconfig have been called
	ASSERT(enabled==2, "Init and SetConfig not yet called?");
	if (enabled!=2)
		return;

	ASSERT(charTransform(' ')==' ', "space is being transformed");
	ASSERT(textScreen, "textScreen is NULL");

	MemSet(textScreen, textScreenSize, ' ');

	modified = 1;
}


static void
ShowObject (UInt16 objectID) 
{
	ASSERT(drawingOK, "Drawing is not ok!");
	if (drawingOK)
	{
		FormType *frm = FrmGetFormPtr(MainFormID);
		ASSERT(frm, "MainFormID is not loaded!");

		if (frm)
		{
			UInt16 objIndex = FrmGetObjectIndex(frm, objectID);
			ASSERT(objIndex!=frmInvalidObjectId, "objID is invalid object id");

			FrmShowObject(frm, objIndex);
		}
	}
}

static void
HideObject (UInt16 objectID) 
{
	ASSERT(drawingOK, "Drawing is not ok!");

	if (drawingOK)
	{
		FormType *frm = FrmGetFormPtr(MainFormID);
		ASSERT(frm, "MainFormID is not loaded!");

		if (frm)
		{
			UInt16 objIndex = FrmGetObjectIndex(frm, objectID);
			ASSERT(objIndex!=frmInvalidObjectId, "objID is invalid object id");

			FrmHideObject(frm, objIndex);
		}
	}
}

void 
Screen_setOutput(UInt8 num, Boolean set)
{
	ASSERT((num>=1)&&(num<=MAX_OUTPUTS), "Output out of range");


	if ((num>=1) && (num<=MAX_OUTPUTS))
	{
		ASSERT(num<sizeof(gpo), "out of range");
		gpo[num-1] = set;
		if ((num==1) && set && AppPrefs.output1Beep)
			SndPlaySystemSound(sndInfo);

		if (drawingOK && AppPrefs.showOutputs)
		{
			FormType *formPtr = FrmGetFormPtr(MainFormID);
			ASSERT(formPtr, "MainForm not loaded");
			if (formPtr)
			{
				UInt16 objIndex = FrmGetObjectIndex(formPtr, Output1 + num-1);

				ASSERT(objIndex!=frmInvalidObjectId, "objID is invalid object id");
				
				FrmSetControlValue(formPtr, objIndex, set);
			}
		}
		// else it will get drawn when drawing is ok
	}
}



static void
showOutputs(void) 
{
	for (UInt32 i=0; i<MAX_OUTPUTS; i++) 
	{
			//CJL:  Remove this stuff.
		if ((i >= 6) || !AppPrefs.showOutputs)
			HideObject(Output1 + i);
		else
			ShowObject(Output1 + i);
	}
}

static void
hideAllKeys (void) 
{
	HideObject(ButtonA);
	HideObject(ButtonB);
	HideObject(ButtonC);
	HideObject(ButtonD);
	HideObject(ButtonE);
	HideObject(ButtonF);
	HideObject(ButtonG);
	HideObject(ButtonH);
	HideObject(ButtonI);
	HideObject(ButtonJ);
	HideObject(ButtonK);
	HideObject(ButtonL);
	HideObject(ButtonM);
	HideObject(ButtonN);
	HideObject(ButtonO);
	HideObject(ButtonP);
	HideObject(ButtonQ);
	HideObject(ButtonR);
	HideObject(ButtonS);
	HideObject(ButtonT);
	HideObject(ButtonU);
	HideObject(ButtonV);
	HideObject(ButtonW);
	HideObject(ButtonX);
#if !NEWSCREEN
	HideObject(ButtonY);
#endif
}

static void
ShowAllKeys (void) 
{
	ShowObject(ButtonA);
	ShowObject(ButtonB);
	ShowObject(ButtonC);
	ShowObject(ButtonD);
	ShowObject(ButtonE);
	ShowObject(ButtonF);
	ShowObject(ButtonG);
	ShowObject(ButtonH);
	ShowObject(ButtonI);
	ShowObject(ButtonJ);
	ShowObject(ButtonK);
	ShowObject(ButtonL);
	ShowObject(ButtonM);
	ShowObject(ButtonN);
	ShowObject(ButtonO);
	ShowObject(ButtonP);
	ShowObject(ButtonQ);
	ShowObject(ButtonR);
	ShowObject(ButtonS);
	ShowObject(ButtonT);
	ShowObject(ButtonU);
	ShowObject(ButtonV);
	ShowObject(ButtonW);
	ShowObject(ButtonX);
#if !NEWSCREEN
	ShowObject(ButtonY);
#endif
}

static void
Show12KeyPad (void) 
{
	ShowObject(ButtonA);
	ShowObject(ButtonB);
	ShowObject(ButtonC);

	ShowObject(ButtonF);
	ShowObject(ButtonG);
	ShowObject(ButtonH);

	ShowObject(ButtonK);
	ShowObject(ButtonL);
	ShowObject(ButtonM);

	ShowObject(ButtonP);
	ShowObject(ButtonQ);
	ShowObject(ButtonR);

	HideObject(ButtonD);
	HideObject(ButtonE);
	HideObject(ButtonI);
	HideObject(ButtonJ);
	HideObject(ButtonN);
	HideObject(ButtonO);
	HideObject(ButtonS);
	HideObject(ButtonT);
	HideObject(ButtonU);
	HideObject(ButtonV);
	HideObject(ButtonW);
	HideObject(ButtonX);
#if !NEWSCREEN
	HideObject(ButtonY);
#endif
}

static void
screen_ShowKeys(void) 
{
	if (!AppPrefs.showKeyboard)
	{
		hideAllKeys();
		return;
	}

	switch(numKeys) 
	{
		case 0:
			hideAllKeys();
			break;
		case 12:
			Show12KeyPad();
			break;
		default:
			ShowAllKeys();
			break;
	}
}

static void
defineChar(UInt16 start_x, UInt16 start_y, const UInt8 *str)
{
	UInt16 x,y, i,j; // UInt16 defchar;
	UInt16 line;
	UInt16 offset_y;
	WinHandle oldwin;
	RectangleType erect; 

	ASSERT(Screen_Fontwin, "Screen_Fontwin is NULL");
	oldwin = WinSetDrawWindow (Screen_Fontwin);

	erect.topLeft.x = start_x;
	erect.topLeft.y = start_y;
	erect.extent.x = FONTCELL_WIDTH;
	erect.extent.y = FONTCELL_HEIGHT;

	//myPushDrawState();
	IndexedColorType oldb,oldf;
	oldb=mySetColor(COLOR_BACKGROUND, colorCharBackIndex);
	oldf=mySetColor(COLOR_FOREGROUND, colorCharForeIndex);

	WinEraseRectangle (&erect, 0);

	for (line = 0; line < FONTCELL_HEIGHT; line++) 
	{
		UInt8 data = *str;

		data = data & 0x1F; // strip unused bits...

		// is there any thing to draw...
		if (data != 0)
		{
			offset_y = start_y;

			// scan across each bit in chr
			for (i = 1, j = FONTCELL_WIDTH - 1; i < 0x40; i <<= 1, j--) 
			{
				x = (start_x + j);
				y = offset_y + line;

				if (data & i)
				{
					if (noWinDrawPixel)
					{
						// Before palmOS 3.5 there was no way to
						// draw a pixel.
						WinDrawLine(x,y,x,y);
					}
					else
					{
						WinDrawPixel(x, y);  //This requires OS 3.5!!
					}
				}
			}
		}
		str++;
	}

	mySetColor(COLOR_BACKGROUND, oldb);
	mySetColor(COLOR_FOREGROUND, oldf);
	//myPopDrawState();
	WinSetDrawWindow (oldwin);
}

// Define chars from top to bottom...
void
Screen_DefineChar (UInt8 ch, const UInt8 *str) 
{

	ASSERT(ch<NUM_SPECIAL_CHARS, "special char out of range");
	
	defineChar(SCHAR_FW_OFFSET_X, SCHAR_FW_OFFSET_Y(ch), str);
	// Save it so we can restore if the color mode is changed.
	if (Screen_ColorPossible)
		MemMove(customCharData[ch], str, FONTCELL_HEIGHT);

	modified=1;
	customNew[ch]=1;
}


#define FULL_BLOCK	'\xFF'
#define BLANK		' '
#define BOTTOM_HALF	'\x1'
#define TOP_HALF	'\x0'
#define LEFT_HALF	'\x2'
#define RIGHT_HALF	'\x3'
#define TOP_RIGHT	'\x6'
#define TOP_LEFT	'\x4'
#define BOTTOM_RIGHT	'\x5'
#define BOTTOM_LEFT	'\x7'


void
Screen_PutNCharsAt (Char *str, UInt16 n, UInt16 x, UInt16 y) 
{
	for (UInt32 i = x; ((i - x) < n) && (i < lcdWidth); i++) 
	{
		Screen_PutChar(str[i-x], i, y);
	}
	//modified = 1;
}

void
Screen_PutChar (UInt8 chr, UInt16 x, UInt16 y) 
{
	ASSERT(enabled==2, "Init and SetConfig not yet called?");
	if (enabled!=2)
		return;

	ASSERT(textScreen, "textScreen is null");
	UInt8 *p = textScreen + (x+y*lcdWidth);

	// transform chr to take our font into account.
	chr = charTransform(chr);

	ASSERT(x<lcdWidth, "x>=w");
	ASSERT(y<lcdHeight, "y>=h");

	ASSERT(textScreen, "textScreen is NULL");

	if (*p != chr) 
	{
		*p = chr;
		modified = 1;
	}
}

static void 
setOutputs(void) 
{
	if (drawingOK && AppPrefs.showOutputs)
	{
		FormType *formPtr = FrmGetFormPtr(MainFormID);
		ASSERT(formPtr, "MainForm not loaded");
		if (!formPtr)
			return;

		for (UInt32 out = 1; out <= MAX_OUTPUTS; out++)
		{
			UInt16 objIndex = FrmGetObjectIndex(formPtr, Output1 + out-1);
			ASSERT(objIndex!=frmInvalidObjectId, "objID is invalid object id");
			ControlType *control = (ControlType *)FrmGetObjectPtr(formPtr,
				   										objIndex);
			ASSERT(control, "control is NULL");

			CtlSetValue(control, gpo[out-1]);
		}
	}
}
