/******************************************************************************
 *
 *  PalmOrb - Emulate a Matrix Orbital LCD Display with Keypad
 *  Copyright (C) 2004  Chris Lansley  <palmorb@lansley.me.uk>
 *  Copyright (C) 2001  David Douthitt 
 *  Copyright (C) 2000  J Robert Ray
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  $Source: /cvsroot/palmorb/palmorb/serial.cpp,v $
 *  $Revision: 1.12 $ $Date: 2004/11/02 21:03:46 $
 *
 ******************************************************************************/

    /***************************************************************/
    /* This file contains mostly the code for handing serial port. */
    /***************************************************************/

#include <PalmOS.h>
#include <SerialMgrOld.h>

#include "palmorb.h"
#include "serial.h"
#include "debug.h"
#include "types.h"

#define min(x,y) ((x) < (y) ? (x) : (y))
#define DEFAULT_BAUD_RATE 19200
#define CUSTOM_SERIAL_QUEUE_SIZE 2048

/* Exported Functions */
Err Serial_Init(void);
void Serial_Fini(void);

Err Serial_OpenPort(void);
void Serial_ClosePort(void);

UInt32 Serial_ReadPort(UInt8 **readBuffer);
void Serial_CloseData(UInt8 *readBuffer, UInt32 numBytes);
void Serial_SendChar(Char c);
void Serial_SetBaudRate(UInt32 baudrate);
UInt32 Serial_GetBaudRate(void);

/* Exported variables */
Boolean Serial_newMan = false;
  // Serial stats:
UInt32 Serial_LineErrorParity = 0;
UInt32 Serial_LineErrorHWOverrun = 0;
UInt32 Serial_LineErrorFraming = 0;
UInt32 Serial_LineErrorBreak = 0;
UInt32 Serial_LineErrorHShake = 0;
UInt32 Serial_LineErrorSWOverrun = 0;
UInt32 Serial_LineErrorCarrierLost = 0;
UInt32 Serial_NonLineErrors = 0;
UInt32 Serial_Errors = 0;
UInt32 Serial_minSize =0xFFFFFFFF;
UInt32 Serial_maxSize =0;
UInt32 Serial_lastSize=0;
UInt32 Serial_sizeTotal=0;
UInt32 Serial_sizeCount=0;
UInt32 Serial_bytesPerSec=0;
UInt32 Serial_maxBps=0;

/* File global varables */
static Boolean connected = false;
static UInt16 portid = 0;
static void *customSerQP = NULL;	// Custom serial queue
static UInt16 serRefNum = 0;

/* local functions */
#if DEBUG_SERIAL
static void debug_serialError(UInt32 error);
#endif // DEBUG_SERIAL

void
debug_sendBuf(const UInt8 *buf, UInt32 bytes)
{
	Err err = 0;

	ASSERT(connected, "not connected!");

	if (Serial_newMan)
		SrmSend(portid, buf, bytes, &err);
	else
		SerSend(serRefNum, buf, bytes, &err);


	if (err != errNone)
	{
		SERIALASSERT(0, "Error sending");
		if (err == serErrLineErr)
		{
			if (Serial_newMan)
				SrmClearErr(portid);
			else
				SerClearErr(serRefNum);
		}
		FrmAlert(ErrTransmittingAlert);
	}
}




Err
Serial_Init(void)
{
	// Check if the new serial manager is avaiable.
	UInt32 avaiable = 0;
	FtrGet(sysFileCSerialMgr, sysFtrNewSerialPresent, &avaiable);
	if (avaiable)
		Serial_newMan = true;

	return errNone;
}

void
Serial_Fini(void)
{
	ASSERT(portid == 0, "port not closed");
	ASSERT(serRefNum == 0, "port not closed");
	ASSERT(connected == 0, "port not closed");
}

static UInt32
serial_convertflags(void)
{
	UInt32 flags = 0;

	// Ensure that new serial manager and old serial manager use same values
	// for their defines. Otherwise code below is broken.
	ASSERT((srmSettingsFlagStopBits1 == serSettingsFlagStopBits1) &&
		   (srmSettingsFlagStopBits2 == serSettingsFlagStopBits2) &&
		   (srmSettingsFlagParityOnM == serSettingsFlagParityOnM) &&
		   (srmSettingsFlagParityEvenM == serSettingsFlagParityEvenM) &&
		   (srmSettingsFlagRTSAutoM == serSettingsFlagRTSAutoM) &&
		   (srmSettingsFlagCTSAutoM == serSettingsFlagCTSAutoM) &&
		   (srmSettingsFlagBitsPerChar5 == serSettingsFlagBitsPerChar5) &&
		   (srmSettingsFlagBitsPerChar6 == serSettingsFlagBitsPerChar6) &&
		   (srmSettingsFlagBitsPerChar7 == serSettingsFlagBitsPerChar7) &&
		   (srmSettingsFlagBitsPerChar8 == serSettingsFlagBitsPerChar8),
		   "Mismatch between old and new serial manager defs");

	/* 
	 * Convert user prefs into serial flags
	 */

	// Stop bits
	if (AppPrefs.serialStopBits == 1)
		flags |= srmSettingsFlagStopBits1;
	else if (AppPrefs.serialStopBits == 2)
		flags |= srmSettingsFlagStopBits2;
	else
		ASSERT(0, "invalid stop bits");

	// Parity
	switch (AppPrefs.serialParity)
	{
		case SERIAL_PARITY_NONE:
			break;

		case SERIAL_PARITY_ODD:
			flags |= srmSettingsFlagParityOnM;
			break;

		case SERIAL_PARITY_EVEN:
			flags |= srmSettingsFlagParityOnM | srmSettingsFlagParityEvenM;
			break;

		default:
			ASSERT(0, "invalid parity");

	}

	// Data bits
	switch (AppPrefs.serialDataBits)
	{
		case 5:
			flags |= srmSettingsFlagBitsPerChar5;
			break;
		case 6:
			flags |= srmSettingsFlagBitsPerChar6;
			break;
		case 7:
			flags |= srmSettingsFlagBitsPerChar7;
			break;
		case 8:
			flags |= srmSettingsFlagBitsPerChar8;
			break;
		default:
			ASSERT(0, "invalid number of data bits");
	}

	if (AppPrefs.serialFlowCtl)
	{
		flags |= srmSettingsFlagRTSAutoM;

		if (Serial_newMan)
			flags |= srmSettingsFlagFlowControlIn;
	}

	return flags;
}

Err
Serial_OpenPort(void)
{
	Err error = 0;
	UInt16 size;

	UInt32 flags = serial_convertflags();	// convert user prefs to serial flags

	if (Serial_newMan)
	{
		UInt32 port;

		switch (AppPrefs.serialDev)
		{
			case SERIAL_DEV_AUTO: port=serPortCradlePort; break;
			case SERIAL_DEV_RS232: port=serPortCradleRS232Port; break;
			case SERIAL_DEV_USB: port=serPortCradleUSBPort; break;
			case SERIAL_DEV_IR: port=sysFileCVirtIrComm; break;
			case SERIAL_DEV_BT: port=sysFileCVirtRfComm; break;
			default:
				ASSERT(0, "unknown serial device");
				port=serPortCradlePort; 
				break;
		}

		error = SrmOpen(port, AppPrefs.serialBaud, &portid);
	}
	else
	{
		// Find serial library
		error = SysLibFind("Serial Library", &serRefNum);
		if (error)
		{
			ASSERT(0, "can't find serial library");
			serRefNum = 0;
			return (error);
		}

		if (serRefNum)
		{
			error = SerOpen(serRefNum, 0, AppPrefs.serialBaud);
		}
	}

	switch (error)
	{
		case errNone:
			break;

		case serErrAlreadyOpen:
			//ASSERT(0, "serErrAlreadyOpen");
			if (Serial_newMan)
				SrmClose(portid);
			else
				SerClose(serRefNum);
			FrmAlert(ErrSerialBusyAlert);
			return error;
			break;

		default:
			ASSERT1(0, "Error opening port: %x", error);
			FrmAlert(ErrOpeningPortAlert);
			return error;
			break;
	}

	connected = true;


	if (Serial_newMan)
	{
		UInt32 baud = AppPrefs.serialBaud;

		// Set default baud rate
		size = sizeof(baud);

		error = SrmControl(portid, srmCtlSetBaudRate, &baud, &size);
		ASSERT(error == errNone, "Failed to set baud rate");
	}

	// configure the serial port

	if (Serial_newMan)
	{
		size = sizeof(flags);

		error = SrmControl(portid, srmCtlSetFlags, &flags, &size);
	}
	else
	{
		SerSettingsType set;

		set.baudRate = AppPrefs.serialBaud;
		set.flags = flags;
		set.ctsTimeout = serDefaultCTSTimeout;

		error = SerSetSettings(serRefNum, &set);
	}

	if (error != errNone)
	{
		FrmAlert(SerialConfigAlert);
	}


	if (AppPrefs.serialBigBuf)
	{
		// Add a custom queue - the default one is 512 bytes.
		// We really don't want to lose serial data - so increase it
		// just in case. No error checking in this code 

		// Allocate a dynamic memory chunk for our custom receive queue. 
		customSerQP = MemPtrNew(CUSTOM_SERIAL_QUEUE_SIZE);
		ASSERT(customSerQP, "failed to allocate queue");
		// Replace the default receive queue. 
		if (customSerQP)
		{
			Err err;	// use our own to avoid set the returned error

			if (Serial_newMan)
			{
				err = SrmSetReceiveBuffer(portid, customSerQP,
										  CUSTOM_SERIAL_QUEUE_SIZE);
			}
			else
			{
				err = SerSetReceiveBuffer(serRefNum, customSerQP,
										  CUSTOM_SERIAL_QUEUE_SIZE);
			}
			ASSERT(err == errNone, "SetReceiveBuffer errored");
			// if errors - we will just have to use the default queue.
			// but lets free the memory up
			if (err != errNone)
			{
				//stop using our queue
				if (Serial_newMan)
					err = SrmSetReceiveBuffer(portid, NULL, 0);
				else
					err = SerSetReceiveBuffer(serRefNum, NULL, 0);

				ASSERT(err == errNone, "SetReceiveBuffer errored");
				MemPtrFree(customSerQP);
				customSerQP = NULL;
			}
		}
	}

	// ignore the above errors - the main one has been
	// handled; not being able to open the port.
	// With the others just continue...
	error = errNone;

	return error;
}

void
Serial_ClosePort(void)
{
	Err error = 0;

	if (!connected)
		return;

	if (customSerQP)
	{
		if (Serial_newMan)
			error = SrmSetReceiveBuffer(portid, NULL, 0);	//stop using our queue
		else
			error = SerSetReceiveBuffer(serRefNum, NULL, 0);	//stop using our q

		ASSERT(error == errNone, "SetReceiveBuffer errored");
		MemPtrFree(customSerQP);
		customSerQP = NULL;
	}

	if (Serial_newMan && portid)
		SrmClose(portid);
	else if ((!Serial_newMan) && serRefNum)
		SerClose(serRefNum);

	portid = 0;
	serRefNum = 0;
	connected = 0;
}



void
Serial_SendChar(UInt8 c)
{
	Err err = 0;

	if (!connected)
		return;

	if (Serial_newMan)
		SrmSend(portid, &c, 1, &err);
	else
		SerSend(serRefNum, &c, 1, &err);

	if (err != errNone)
	{
		SERIALASSERT(0, "Error sending");
		if (err == serErrLineErr)
		{
			if (Serial_newMan)
				SrmClearErr(portid);
			else
				SerClearErr(serRefNum);
		}
		FrmAlert(ErrTransmittingAlert);
	}
}

UInt32
Serial_GetBaudRate(void)
{
	Err err;
	UInt16 size;
	UInt32 baud = 0;

	if (!connected)
		return 0;

	size = sizeof(Int32);

	if (Serial_newMan)
	{
		err = SrmControl(portid, srmCtlGetBaudRate, &baud, &size);
		ASSERT(err == errNone, "GetBaudRate failed");
	}
	else
	{
		SerSettingsType settings;

		err = SerGetSettings(serRefNum, &settings);
		ASSERT(err == errNone, "SerGetSettings failed");
		if (err != errNone)
			return 0;
		baud = settings.baudRate;
	}
	return baud;
}

void
Serial_SetBaudRate(UInt32 baudrate)
{
	Err err;

	ASSERT(connected, "not connected!");

	if (Serial_newMan)
	{
		UInt16 size;
		size = sizeof(Int32);
		err = SrmControl(portid, srmCtlSetBaudRate, &baudrate, &size);
		ASSERT(err == errNone, "setbaudrate failed");
	}
	else
	{
		SerSettingsType settings;

		err = SerGetSettings(serRefNum, &settings);
		ASSERT(err == errNone, "SerGetSettings failed");

		if (err != errNone)
		{
			settings.baudRate = baudrate;
			err = SerSetSettings(serRefNum, &settings);
			ASSERT(err == errNone, "SerSetSettings failed");
		}
	}
}


#if DEBUG_SERIAL
static void
debug_serialError(UInt32 error)
{
	switch (error)
	{
		case serErrBadPort:
			SERIALASSERT(0, "debug_serialError: serErrBadPort");
			break;
		case serErrNotOpen:
			SERIALASSERT(0, "debug_serialError: serErrNotOpen");
			break;
		case serErrLineErr:
			SERIALASSERT(0, "debug_serialError: serErrLineErr");
			UInt32 lineStatus;
			UInt16 lineErrors;
			Err err;
			if (Serial_newMan)
				err = SrmGetStatus(portid, &lineStatus, &lineErrors);
			else
			{
				Boolean ctsOn, dsrOn;
				lineErrors = SerGetStatus(serRefNum, &ctsOn, &dsrOn);
				lineStatus = (ctsOn ? srmStatusCtsOn : 0) |
								(dsrOn ? srmStatusDsrOn : 0);
			}

			if (err == errNone)
			{
				Char stat[25], er[140];
				StrPrintF(stat, "%s%s%s%s",
						  (lineStatus & srmStatusCtsOn ? "|CTS|" : ""),
						  (lineStatus & srmStatusRtsOn ? "|RTS|" : ""),
						  (lineStatus & srmStatusDsrOn ? "|DSR|" : ""),
						  (lineStatus & srmStatusBreakSigOn ? "|BRK|" : ""));
				StrPrintF(er, "%s:%s%s%s%s%s%s%s",
						  stat,
						  (lineErrors & serLineErrorParity ? "|PAR|" : ""),
						  (lineErrors & serLineErrorHWOverrun ? "|HOv|" : ""),
						  (lineErrors & serLineErrorFraming ? "|Frm|" : ""),
						  (lineErrors & serLineErrorBreak ? "|Brk|" : ""),
						  (lineErrors & serLineErrorHShake ? "|Hsk|" : ""),
						  (lineErrors & serLineErrorSWOverrun ? "|SOv|" : ""),
						  (lineErrors & serLineErrorCarrierLost ? "|NOc|" : ""));
				SERIALASSERT(0, er);
			}
			break;
		default:
			SERIALASSERT(0, "debug_serialError: errored");
	}
}
#endif // DEBUG_SERIAL

UInt32
Serial_ReadPort(UInt8 **readBuffer)
{
	UInt32 numBytes = 0;
	Err error;

	if (!connected)
		return (0);

	if (Serial_newMan)
		error = SrmReceiveWindowOpen(portid, readBuffer, &numBytes);
	else
		error = SerReceiveWindowOpen(serRefNum, readBuffer, &numBytes);

	ASSERT1(numBytes <= CUSTOM_SERIAL_QUEUE_SIZE, "too many bytes %u",
			numBytes);

	if (error)
	{
		if (error == serErrLineErr)
		{
			/* Keep error stats, if req */
			if (AppPrefs.serialErrDetails)
			{
				UInt32 lineStatus;
				UInt16 lineErrors;
				Err err = errNone;

				if (Serial_newMan)
					err = SrmGetStatus(portid, &lineStatus, &lineErrors);
				else
				{
					Boolean ctsOn, dsrOn;
					lineErrors = SerGetStatus(serRefNum, &ctsOn, &dsrOn);
					lineStatus = (ctsOn ? srmStatusCtsOn : 0) |
									(dsrOn ? srmStatusDsrOn : 0);
				}

				if (err == errNone)
				{
					if (lineErrors & serLineErrorParity)
						Serial_LineErrorParity++;
					if (lineErrors & serLineErrorHWOverrun)
						Serial_LineErrorHWOverrun++;
					if (lineErrors & serLineErrorFraming)
						Serial_LineErrorFraming++;
					if (lineErrors & serLineErrorBreak)
						Serial_LineErrorBreak++;
					if (lineErrors & serLineErrorHShake)
						Serial_LineErrorBreak++;
					if (lineErrors & serLineErrorSWOverrun)
						Serial_LineErrorSWOverrun++;
					if (lineErrors & serLineErrorCarrierLost)
						Serial_LineErrorCarrierLost++;
				}
			}

			if (Serial_newMan)
				SrmClearErr(portid);
			else
				SerClearErr(serRefNum);
		}
		else
		{
			Serial_NonLineErrors++;
			ASSERT1(0, "nonline error", error);
#if DEBUG_SERIAL
			debug_serialError(error);
#endif //DEBUG_SERIAL
		}


		/*
		 * Keep track of serial errors.
		 * Display a warning if too many occur.
		 */
		Serial_Errors++;

		if ((Serial_Errors > 100) && (AppPrefs.warnSerialErrors))
		{
			static MInt8 alreadyWarned = 0;

			if (!alreadyWarned)
			{
				alreadyWarned = 1;

				FrmAlert(ErrSerialErrorsAlert);
			}
		}
	}

	if (numBytes)
	{
		/*
		 * Keep some serial stats
		 */

		static UInt32 lasttime = 0;
        static UInt32 lastBytes = 0;
        UInt32 now = TimGetTicks();
        static UInt16 ticksPerSecond = SysTicksPerSecond();

        if (Serial_minSize > numBytes)
            Serial_minSize = numBytes;

        if (Serial_maxSize < numBytes)
            Serial_maxSize = numBytes;

        Serial_sizeCount++;
        Serial_lastSize = numBytes;
        Serial_sizeTotal += numBytes;

        if ((now - lasttime > ticksPerSecond) || (now < lasttime))
        {
            UInt32 seconds = (now - lasttime) / ticksPerSecond;
            Serial_bytesPerSec = (Serial_sizeTotal - lastBytes) / seconds;
            lastBytes = Serial_sizeTotal;
            lasttime = now;
            if (Serial_maxBps < Serial_bytesPerSec)
                Serial_maxBps = Serial_bytesPerSec;
		}
	}
	else
	{
		// I assume we have to Close it even if zero bytes are avail?
		if (Serial_newMan)
			error = SrmReceiveWindowClose(portid, numBytes);
		else
			error = SerReceiveWindowClose(serRefNum, numBytes);
		ASSERT(error == errNone, "SrmReceive errored");
	}

	return numBytes;
}

void
Serial_CloseData(UInt8 *readBuffer, UInt32 numBytes)
{
	Err error;

	if (!connected)
		return;

	if (Serial_newMan)
		error = SrmReceiveWindowClose(portid, numBytes);
	else
		error = SerReceiveWindowClose(serRefNum, numBytes);
	ASSERT(error == errNone, "SrmReceive errored");
}
